from .tbool import TBool, TBoolInst, TBoolSeq, TBoolSeqSet
from .tfloat import TFloat, TFloatInst, TFloatSeq, TFloatSeqSet
from .tint import TInt, TIntInst, TIntSeq, TIntSeqSet
from .tnumber import TNumber
from .tpoint import TPoint, TPointInst, TPointSeq, TPointSeqSet, \
    TGeomPoint, TGeomPointInst, TGeomPointSeq, TGeomPointSeqSet, \
    TGeogPoint, TGeogPointInst, TGeogPointSeq, TGeogPointSeqSet
from .ttext import TText, TTextInst, TTextSeq, TTextSeqSet

__all__ = [
    'TBool', 'TBoolInst', 'TBoolSeq', 'TBoolSeqSet',
    'TNumber',
    'TInt', 'TIntInst', 'TIntSeq', 'TIntSeqSet',
    'TFloat', 'TFloatInst', 'TFloatSeq', 'TFloatSeqSet',
    'TText', 'TTextInst', 'TTextSeq', 'TTextSeqSet',
    'TPoint', 'TPointInst', 'TPointSeq', 'TPointSeqSet',
    'TGeomPoint', 'TGeomPointInst', 'TGeomPointSeq', 'TGeomPointSeqSet',
    'TGeogPoint', 'TGeogPointInst', 'TGeogPointSeq', 'TGeogPointSeqSet']



from __future__ import annotations

from typing import Union, overload, Optional, TYPE_CHECKING

from pymeos_cffi import intersection_floatspan_float, distance_floatspan_float, \
    floatspan_in, floatspan_lower, floatspan_upper, floatspan_shift_scale, \
    contains_floatspan_float, adjacent_floatspan_float, \
    float_to_floatspan, span_eq, \
    left_floatspan_float, overleft_floatspan_float, \
    right_floatspan_float, overright_floatspan_float, \
    intersection_span_span, intersection_spanset_span, \
    minus_floatspan_float, minus_span_span, minus_spanset_span, \
    union_floatspan_float, union_span_span, union_spanset_span, \
    floatspan_out, floatspan_make, span_width, floatspan_intspan

from .. import Span

if TYPE_CHECKING:
    from .floatspanset import FloatSpanSet
    from .intspan import IntSpan


class FloatSpan(Span[float]):
    """
    Class for representing sets of contiguous float values between a lower and
    an upper bound. The bounds may be inclusive or not.

    ``FloatSpan`` objects can be created with a single argument of type string
    as in MobilityDB.

        >>> FloatSpan('(2.5, 5.21]')

    Another possibility is to provide the ``lower`` and ``upper`` named
    parameters (of type str or float), and optionally indicate whether the
    bounds are inclusive or exclusive (by default, the lower bound is inclusive
    and the upper is exclusive):

        >>> FloatSpan(lower=2.0, upper=5.8)
        >>> FloatSpan(lower=2.0, upper=5.8, lower_inc=False, upper_inc=True)
        >>> FloatSpan(lower='2.0', upper='5.8', upper_inc=True)
    """

    __slots__ = ['_inner']

    _mobilitydb_name = 'floatspan'

    _parse_function = floatspan_in
    _parse_value_function = float
    _make_function = floatspan_make

    # ------------------------- Output ----------------------------------------
    def __str__(self, max_decimals: int = 15) -> str:
        """
        Return the string representation of the content of ``self``.

        Returns:
            A new :class:`str` instance

        MEOS Functions:
            floatspan_out
        """
        return floatspan_out(self._inner, max_decimals)

    # ------------------------- Conversions -----------------------------------

    def to_spanset(self) -> FloatSpanSet:
        """
        Returns a spanset containing ``self``.

        Returns:
            A new :class:`FloatSpanSet` instance

        MEOS Functions:
            span_to_spanset
        """
        from .floatspanset import FloatSpanSet
        return FloatSpanSet(_inner=super().to_spanset())

    def to_intspan(self) -> IntSpan:
        """
        Converts ``self`` to a :class:`IntSpan` instance.

        Returns:
            A new :class:`IntSpan` instance

        MEOS Functions:
            floatspan_intspan
        """
        from .intspan import IntSpan
        return IntSpan(_inner=floatspan_intspan(self._inner))

    # ------------------------- Accessors -------------------------------------
    def lower(self) -> float:
        """
        Returns the lower bound of ``self``.

        Returns:
            The lower bound of the span as a :class:`float`

        MEOS Functions:
            period_lower
        """

        return floatspan_lower(self._inner)

    def upper(self) -> float:
        """
        Returns the upper bound of ``self``.

        Returns:
            The upper bound of the span as a :class:`float`

        MEOS Functions:
            period_upper
        """
        return floatspan_upper(self._inner)

    def width(self) -> float:
        """
        Returns the width of ``self``.

        Returns:
            Returns a `float` representing the width of the span

        MEOS Functions:
            span_width
        """
        return span_width(self._inner)

    # ------------------------- Transformations -------------------------------
    def shift(self, delta: float) -> FloatSpan:
        """
        Return a new ``FloatSpan`` with the lower and upper bounds shifted by
        ``delta``.

        Args:
            delta: The value to shift by

        Returns:
            A new ``FloatSpan`` instance

        MEOS Functions:
            floatspan_shift_scale
        """
        return self.shift_scale(delta, None)

    def scale(self, width: float) -> FloatSpan:
        """
        Return a new ``FloatSpan`` with the lower and upper bounds scaled so
        that the width is ``width``.

        Args:
            width: The new width

        Returns:
            A new ``FloatSpan`` instance

        MEOS Functions:
            floatspan_shift_scale
        """
        return self.shift_scale(None, width)

    def shift_scale(self, delta: Optional[float], width: Optional[float]) -> FloatSpan:
        """
        Return a new ``FloatSpan`` with the lower and upper bounds shifted by
        ``delta`` and scaled so that the width is ``width``.

        Args:
            delta: The value to shift by
            width: The new width

        Returns:
            A new ``FloatSpan`` instance

        MEOS Functions:
            floatspan_shift_scale
        """
        d = delta if delta is not None else 0
        w = width if width is not None else 0
        modified = floatspan_shift_scale(self._inner, d, w, delta is not None,
                                         width is not None)
        return FloatSpan(_inner=modified)

    # ------------------------- Topological Operations --------------------------------

    def is_adjacent(self, other: Union[int, float, FloatSpan, FloatSpanSet]) -> bool:
        """
        Returns whether ``self`` is adjacent to ``other``. That is, they share
        a bound but only one of them contains it.

        Args:
            other: object to compare with

        Returns:
            True if adjacent, False otherwise

        MEOS Functions:
            adjacent_span_span, adjacent_span_spanset, adjacent_floatspan_float
        """
        if isinstance(other, int) or isinstance(other, float):
            return adjacent_floatspan_float(self._inner, float(other))
        else:
            return super().is_adjacent(other)

    def contains(self, content: Union[int, float, FloatSpan, FloatSpanSet]) -> bool:
        """
        Returns whether ``self`` contains ``content``.

        Args:
            content: object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_set_set, contains_floatspan_float
        """
        if isinstance(content, int) or isinstance(content, float):
            return contains_floatspan_float(self._inner, float(content))
        else:
            return super().contains(content)

    def is_same(self, other: Union[int, float, FloatSpan, FloatSpanSet]) -> bool:
        """
        Returns whether ``self`` and the bounding period of ``other`` is the
        same.

        Args:
            other: object to compare with

        Returns:
            True if equal, False otherwise

        MEOS Functions:
            same_period_temporal
        """
        if isinstance(other, int) or isinstance(other, float):
            return span_eq(self._inner, float_to_floatspan(float(other)))
        else:
            return super().is_same(other)

    # ------------------------- Position Operations ---------------------------
    def is_left(self, other: Union[int, float, FloatSpan, FloatSpanSet]) -> bool:
        """
        Returns whether ``self`` is strictly before ``other``. That is,
        ``self`` ends before ``other`` starts.

        Args:
            other: object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
            left_span_span, left_span_spanset, left_floatspan_float
        """
        if isinstance(other, int) or isinstance(other, float):
            return left_floatspan_float(self._inner, float(other))
        else:
            return super().is_left(other)

    def is_over_or_left(self, other: Union[int, float, FloatSpan, FloatSpanSet]) -> bool:
        """
        Returns whether ``self`` is before ``other`` allowing overlap. That is,
        ``self`` ends before ``other`` ends (or at the same value).

        Args:
            other: object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
            overleft_span_span, overleft_span_spanset, overleft_floatspan_float
        """
        if isinstance(other, int) or isinstance(other, float):
            return overleft_floatspan_float(self._inner, float(other))
        else:
            return super().is_over_or_left(other)

    def is_right(self, other: Union[int, float, FloatSpan, FloatSpanSet]) -> bool:
        """
        Returns whether ``self`` is strictly after ``other``. That is, ``self``
        starts after ``other`` ends.

        Args:
            other: object to compare with

        Returns:
            True if after, False otherwise

        MEOS Functions:
            right_span_span, right_span_spanset, right_floatspan_float
        """
        if isinstance(other, int) or isinstance(other, float):
            return right_floatspan_float(self._inner, float(other))
        else:
            return super().is_right(other)

    def is_over_or_right(self, other: Union[float, FloatSpan, FloatSpanSet]) -> bool:
        """
        Returns whether ``self`` is after ``other`` allowing overlap. That is,
        ``self`` starts after ``other`` starts (or at the same value).

        Args:
            other: object to compare with

        Returns:
            True if overlapping or after, False otherwise

        MEOS Functions:
            overright_span_span, overright_span_spanset, overright_floatspan_float
        """
        if isinstance(other, int) or isinstance(other, float):
            return overright_floatspan_float(self._inner, float(other))
        else:
            return super().is_over_or_right(other)

    # ------------------------- Distance Operations ---------------------------
    def distance(self, other: Union[int, float, FloatSpan, FloatSpanSet]) -> float:
        """
        Returns the distance between ``self`` and ``other``.

        Args:
            other: object to compare with

        Returns:
            A float value

        MEOS Functions:
            distance_span_span, distance_span_spanset, distance_floatspan_float,
        """
        if isinstance(other, int) or isinstance(other, float):
            return distance_floatspan_float(self._inner, float(other))
        else:
            return super().distance(other)

    # ------------------------- Set Operations --------------------------------
    @overload
    def intersection(self, other: Union[int, float]) -> Optional[float]:
        ...

    @overload
    def intersection(self, other: FloatSpan) -> Optional[FloatSpan]:
        ...

    @overload
    def intersection(self, other: FloatSpanSet) -> Optional[FloatSpanSet]:
        ...

    def intersection(self, other):
        """
        Returns the intersection of ``self`` and ``other``.

        Args:
            other: object to intersect with

        Returns:
            A :class:`Collection[float]` instance. The actual class depends on
            ``other``.

        MEOS Functions:
            intersection_span_span, intersection_spanset_span,
            intersection_floatset_float
        """
        from .floatspanset import FloatSpanSet
        if isinstance(other, int) or isinstance(other, float):
            return intersection_floatspan_float(self._inner, float(other))
        elif isinstance(other, FloatSpan):
            result = intersection_span_span(self._inner, other._inner)
            return FloatSpan(_inner=result) if result is not None else None
        elif isinstance(other, FloatSpanSet):
            result = intersection_spanset_span(other._inner, self._inner)
            return FloatSpanSet(_inner=result) if result is not None else None
        else:
            super().intersection(other)

    def minus(self, other: Union[int, float, FloatSpan, FloatSpanSet]) -> FloatSpanSet:
        """
        Returns the difference of ``self`` and ``other``.

        Args:
            other: object to diff with

        Returns:
            A :class:`FloatSpanSet` instance.

        MEOS Functions:
            minus_span_span, minus_spanset_span, minus_floatspan_float
        """
        from .floatspanset import FloatSpanSet
        if isinstance(other, int) or isinstance(other, float):
            result = minus_floatspan_float(self._inner, float(other))
        elif isinstance(other, FloatSpan):
            result = minus_span_span(self._inner, other._inner)
        elif isinstance(other, FloatSpanSet):
            result = minus_spanset_span(other._inner, self._inner)
        else:
            result = super().minus(other)
        return FloatSpanSet(_inner=result) if result is not None else None

    def union(self, other: Union[int, float, FloatSpan, FloatSpanSet]) -> FloatSpanSet:
        """
        Returns the union of ``self`` and ``other``.

        Args:
            other: object to merge with

        Returns:
            A :class:`PeriodSet` instance.

        MEOS Functions:
            union_spanset_span, union_span_span, union_floatspan_float
        """
        from .floatspanset import FloatSpanSet
        if isinstance(other, int) or isinstance(other, float):
            result = union_floatspan_float(self._inner, float(other))
        elif isinstance(other, FloatSpan):
            result = union_span_span(self._inner, other._inner)
        elif isinstance(other, FloatSpanSet):
            result = union_spanset_span(other._inner, self._inner)
        else:
            result = super().union(other)
        return FloatSpanSet(_inner=result) if result is not None else None



from __future__ import annotations

from datetime import datetime, timedelta
from typing import Optional, Union, overload, TYPE_CHECKING, get_args

from dateutil.parser import parse
from pymeos_cffi import *

from .time_collection import TimeCollection
from ..base.span import Span

if TYPE_CHECKING:
    from ...temporal import Temporal
    from ...boxes import Box
    from .periodset import PeriodSet
    from .time import Time
    from .timestampset import TimestampSet


class Period(Span[datetime], TimeCollection):
    """
    Class for representing sets of contiguous timestamps between a lower and
    an upper bound. The bounds may be inclusive or not.

    ``Period`` objects can be created with a single argument of type string
    as in MobilityDB.

        >>> Period('(2019-09-08 00:00:00+01, 2019-09-10 00:00:00+01)')

    Another possibility is to provide the ``lower`` and ``upper`` named
    parameters (of type str or datetime), and optionally indicate whether the
    bounds are inclusive or exclusive (by default, the lower bound is inclusive
    and the upper is exclusive):

        >>> Period(lower='2019-09-08 00:00:00+01', upper='2019-09-10 00:00:00+01')
        >>> Period(lower='2019-09-08 00:00:00+01', upper='2019-09-10 00:00:00+01', lower_inc=False, upper_inc=True)
        >>> Period(lower=parse('2019-09-08 00:00:00+01'), upper=parse('2019-09-10 00:00:00+01'), upper_inc=True)
    """

    __slots__ = ['_inner']

    _mobilitydb_name = 'tstzspan'

    _parse_function = period_in
    _parse_value_function = lambda x: pg_timestamptz_in(x, -1) if isinstance(x, str) else datetime_to_timestamptz(x)
    _make_function = period_make

    # ------------------------- Constructors ----------------------------------

    # ------------------------- Output ----------------------------------------
    def __str__(self):
        """
        Return the string representation of the content of ``self``.

        Returns:
            A new :class:`str` instance

        MEOS Functions:
            period_out
        """
        return period_out(self._inner)

    # ------------------------- Conversions -----------------------------------

    def to_spanset(self) -> PeriodSet:
        """
        Returns a period set containing ``self``.

        Returns:
            A new :class:`PeriodSet` instance

        MEOS Functions:
            span_to_spanset
        """
        from .periodset import PeriodSet
        return PeriodSet(_inner=super().to_spanset())

    def to_periodset(self) -> PeriodSet:
        """
        Returns a period set containing ``self``.

        Returns:
            A new :class:`PeriodSet` instance

        MEOS Functions:
            span_to_spanset
        """
        return self.to_spanset()

    # ------------------------- Accessors -------------------------------------
    def lower(self) -> datetime:
        """
        Returns the lower bound of a period

        Returns:
            The lower bound of the period as a :class:`datetime.datetime`

        MEOS Functions:
            period_lower
        """

        return timestamptz_to_datetime(period_lower(self._inner))

    def upper(self) -> datetime:
        """
        Returns the upper bound of a period

        Returns:
            The upper bound of the period as a :class:`datetime.datetime`

        MEOS Functions:
            period_upper
        """
        return timestamptz_to_datetime(period_upper(self._inner))

    def duration(self) -> timedelta:
        """
        Returns the duration of the period.

        Returns:
            A :class:`datetime.timedelta` instance representing the duration of
            the period

        MEOS Functions:
            period_duration
        """
        return interval_to_timedelta(period_duration(self._inner))

    def duration_in_seconds(self) -> float:
        """
        Returns the duration of the period.

        Returns:
            Returns a `float` representing the duration of the period in seconds

        MEOS Functions:
            span_width
        """
        return self.width()

    # ------------------------- Transformations -------------------------------
    def shift(self, delta: timedelta) -> Period:
        """
        Returns a new period that is the result of shifting ``self`` by ``delta``

        Examples:
            >>> Period('[2000-01-01, 2000-01-10]').shift(timedelta(days=2))
            >>> 'Period([2000-01-03 00:00:00+01, 2000-01-12 00:00:00+01])'

        Args:
            delta: :class:`datetime.timedelta` instance to shift

        Returns:
            A new :class:`Period` instance

        MEOS Functions:
            period_shift_scale
        """
        return self.shift_scale(shift=delta)

    def scale(self, duration: timedelta) -> Period:
        """
        Returns a new period that starts as ``self`` but has duration ``duration``

        Examples:
            >>> Period('[2000-01-01, 2000-01-10]').scale(timedelta(days=2))
            >>> 'Period([2000-01-01 00:00:00+01, 2000-01-03 00:00:00+01])'

        Args:
            duration: :class:`datetime.timedelta` instance representing the
            duration of the new period

        Returns:
            A new :class:`Period` instance

        MEOS Functions:
            period_shift_scale
        """
        return self.shift_scale(duration=duration)

    def shift_scale(self, shift: Optional[timedelta] = None,
        duration: Optional[timedelta] = None) -> Period:
        """
        Returns a new period that starts at ``self`` shifted by ``shift`` and
        has duration ``duration``

        Examples:
            >>> Period('[2000-01-01, 2000-01-10]').shift_scale(shift=timedelta(days=2), duration=timedelta(days=4))
            >>> 'Period([2000-01-03 00:00:00+01, 2000-01-07 00:00:00+01])'

        Args:
            shift: :class:`datetime.timedelta` instance to shift
            duration: :class:`datetime.timedelta` instance representing the
            duration of the new period

        Returns:
            A new :class:`Period` instance

        MEOS Functions:
            period_shift_scale
        """
        assert shift is not None or duration is not None, 'shift and scale deltas must not be both None'
        modified = period_shift_scale(
            self._inner,
            timedelta_to_interval(shift) if shift else None,
            timedelta_to_interval(duration) if duration else None,
        )
        return Period(_inner=modified)

    # ------------------------- Topological Operations ------------------------
    def is_adjacent(self, other: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` is temporally adjacent to ``other``. That is,
        they share a bound but only one of them contains it.

        Examples:
            >>> Period('[2012-01-01, 2012-01-02)').is_adjacent(Period('[2012-01-02, 2012-01-03]'))
            >>> True
            >>> Period('[2012-01-01, 2012-01-02]').is_adjacent(Period('[2012-01-02, 2012-01-03]'))
            >>> False  # Both contain bound
            >>> Period('[2012-01-01, 2012-01-02)').is_adjacent(Period('(2012-01-02, 2012-01-03]'))
            >>> False  # Neither contain bound

        Args:
            other: temporal object to compare with

        Returns:
            True if adjacent, False otherwise

        MEOS Functions:
            adjacent_span_span, adjacent_span_spanset, adjacent_period_timestamp,
            adjacent_period_timestampset, adjacent_period_temporal
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return adjacent_period_timestamp(self._inner,
                datetime_to_timestamptz(other))
        elif isinstance(other, Temporal):
            return adjacent_span_span(self._inner, temporal_to_period(other._inner))
        elif isinstance(other, get_args(Box)):
            return adjacent_span_span(self._inner, other.to_period()._inner)
        else:
            return super().is_adjacent(other)

    def is_contained_in(self, container: Union[Period, PeriodSet, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` is temporally contained in ``container``.

        Examples:
            >>> Period('[2012-01-02, 2012-01-03]').is_contained_in(Period('[2012-01-01, 2012-01-04]'))
            >>> True
            >>> Period('(2012-01-01, 2012-01-02)').is_contained_in(Period('[2012-01-01, 2012-01-02]'))
            >>> True
            >>> Period('[2012-01-01, 2012-01-02]').is_contained_in(Period('(2012-01-01, 2012-01-02)'))
            >>> False

        Args:
            container: temporal object to compare with

        Returns:
            True if contained, False otherwise

        MEOS Functions:
            contained_span_span, contained_span_spanset, contained_period_temporal
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(container, Temporal):
            return self.is_contained_in(container.period())
        elif isinstance(container, Box):
            return self.is_contained_in(container.to_period())
        else:
            return super().is_contained_in(container)

    def contains(self, content: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` temporally contains ``content``.

        Examples:
            >>> Period('[2012-01-01, 2012-01-04]').contains(Period('[2012-01-02, 2012-01-03]'))
            >>> True
            >>> Period('[2012-01-01, 2012-01-02]').contains(Period('(2012-01-01, 2012-01-02)'))
            >>> True
            >>> Period('(2012-01-01, 2012-01-02)').contains(Period('[2012-01-01, 2012-01-02]'))
            >>> False

        Args:
            content: temporal object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_span_span, contains_span_spanset, contains_period_timestamp,
            contains_period_timestampset, contains_period_temporal
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(content, datetime):
            return contains_period_timestamp(self._inner,
                datetime_to_timestamptz(content))
        elif isinstance(content, Temporal):
            return self.contains(content.period())
        elif isinstance(content, get_args(Box)):
            return self.contains(content.to_period())
        else:
            return super().contains(content)

    def overlaps(self, other: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` temporally overlaps ``other``. That is, both
        share at least an instant

        Examples:
            >>> Period('[2012-01-01, 2012-01-02]').overlaps(Period('[2012-01-02, 2012-01-03]'))
            >>> True
            >>> Period('[2012-01-01, 2012-01-02)').overlaps(Period('[2012-01-02, 2012-01-03]'))
            >>> False
            >>> Period('[2012-01-01, 2012-01-02)').overlaps(Period('(2012-01-02, 2012-01-03]'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if overlaps, False otherwise

        MEOS Functions:
            overlaps_span_span, overlaps_span_spanset, 
            overlaps_period_timestampset, overlaps_period_temporal
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return overlaps_span_span(self._inner,
                timestamp_to_period(datetime_to_timestamptz(other)))
        elif isinstance(other, Temporal):
            return self.overlaps(other.period())
        elif isinstance(other, get_args(Box)):
            return self.overlaps(other.to_period())
        else:
            return super().overlaps(other)

    def is_same(self, other: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` and the bounding period of ``other`` is the same.

        Args:
            other: temporal object to compare with

        Returns:
            True if equal, False otherwise

        MEOS Functions:
            same_period_temporal
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, Temporal):
            return self.is_same(other.period())
        elif isinstance(other, get_args(Box)):
            return self.is_same(other.to_period())
        elif isinstance(other, datetime):
            return span_eq(self._inner,
                timestamp_to_period(datetime_to_timestamptz(other)))
        else:
            return super().is_same(other)

    # ------------------------- Position Operations ---------------------------
    def is_left(self, other: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` is strictly before ``other``. That is,
        ``self`` ends before ``other`` starts.

        Examples:
            >>> Period('[2012-01-01, 2012-01-02)').is_left(Period('[2012-01-02, 2012-01-03]'))
            >>> True
            >>> Period('[2012-01-01, 2012-01-02)').is_left(Period('(2012-01-02, 2012-01-03]'))
            >>> True
            >>> Period('[2012-01-01, 2012-01-02]').is_left(Period('[2012-01-02, 2012-01-03]'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
            left_span_span, left_span_spanset, before_period_timestamp,
            before_period_timestampset, before_period_temporal
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return overafter_timestamp_period(datetime_to_timestamptz(other),
                self._inner)
        elif isinstance(other, Temporal):
            return self.is_left(other.period())
        elif isinstance(other, get_args(Box)):
            return self.is_left(other.to_period())
        else:
            return super().is_left(other)

    def is_over_or_left(self, other: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` is before ``other`` allowing overlap. That is,
        ``self`` ends before ``other`` ends (or at the same time).

        Examples:
            >>> Period('[2012-01-01, 2012-01-02)').is_over_or_left(Period('[2012-01-02, 2012-01-03]'))
            >>> True
            >>> Period('[2012-01-01, 2012-01-02]').is_over_or_left(Period('[2012-01-02, 2012-01-03]'))
            >>> True
            >>> Period('[2012-01-03, 2012-01-05]').is_over_or_left(Period('[2012-01-01, 2012-01-04]'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
            overleft_span_span, overleft_span_spanset, overbefore_period_timestamp,
            overbefore_period_timestampset, overbefore_period_temporal
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return overbefore_period_timestamp(self._inner,
                datetime_to_timestamptz(other))
        elif isinstance(other, Temporal):
            return self.is_over_or_left(other.period())
        elif isinstance(other, get_args(Box)):
            return self.is_over_or_left(other.to_period())
        else:
            return super().is_over_or_left(other)

    def is_right(self, other: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` is strictly after ``other``. That is, ``self``
        starts after ``other`` ends.

        Examples:
            >>> Period('[2012-01-02, 2012-01-03]').is_right(Period('[2012-01-01, 2012-01-02)'))
            >>> True
            >>> Period('(2012-01-02, 2012-01-03]').is_right(Period('[2012-01-01, 2012-01-02)'))
            >>> True
            >>> Period('[2012-01-02, 2012-01-03]').is_right(Period('[2012-01-01, 2012-01-02]'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if after, False otherwise

        MEOS Functions:
            right_span_span, right_span_spanset, after_period_timestamp,
            after_period_timestampset, after_period_temporal
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return overbefore_timestamp_period(datetime_to_timestamptz(other),
                self._inner)
        elif isinstance(other, Temporal):
            return self.is_right(other.period())
        elif isinstance(other, get_args(Box)):
            return self.is_right(other.to_period())
        else:
            return super().is_right(other)

    def is_over_or_right(self, other: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` is after ``other`` allowing overlap. That is,
        ``self`` starts after ``other`` starts (or at the same time).

        Examples:
            >>> Period('[2012-01-02, 2012-01-03]').is_over_or_right(Period('[2012-01-01, 2012-01-02)'))
            >>> True
            >>> Period('[2012-01-02, 2012-01-03]').is_over_or_right(Period('[2012-01-01, 2012-01-02]'))
            >>> True
            >>> Period('[2012-01-02, 2012-01-03]').is_over_or_right(Period('[2012-01-01, 2012-01-03]'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if overlapping or after, False otherwise

        MEOS Functions:
            overright_span_span, overright_span_spanset, overafter_period_timestamp,
            overafter_period_timestampset, overafter_period_temporal
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return overafter_period_timestamp(self._inner,
                datetime_to_timestamptz(other))
        elif isinstance(other, Temporal):
            return self.is_over_or_right(other.period())
        elif isinstance(other, get_args(Box)):
            return self.is_over_or_right(other.to_period())
        else:
            return super().is_over_or_right(other)

    # ------------------------- Distance Operations ---------------------------
    def distance(self, other: Union[Time, Box, Temporal]) -> timedelta:
        """
        Returns the temporal distance between ``self`` and ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            A :class:`datetime.timedelta` instance

        MEOS Functions:
            distance_span_span, distance_spanset_span, distance_period_timestamp
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return timedelta(seconds=distance_period_timestamp(self._inner,
                datetime_to_timestamptz(other)))
        elif isinstance(other, Temporal):
            return self.distance(other.period())
        elif isinstance(other, get_args(Box)):
            return self.distance(other.to_period())
        else:
            return timedelta(seconds=super().distance(other))

    # ------------------------- Set Operations --------------------------------
    @overload
    def intersection(self, other: datetime) -> Optional[datetime]:
        ...

    @overload
    def intersection(self, other: Period) -> Optional[Period]:
        ...

    @overload
    def intersection(self, other: Union[TimestampSet, PeriodSet]) -> Optional[PeriodSet]:
        ...

    def intersection(self, other: Time) -> Optional[Time]:
        """
        Returns the temporal intersection of ``self`` and ``other``.

        Args:
            other: temporal object to intersect with

        Returns:
            A :class:`Time` instance. The actual class depends on ``other``.

        MEOS Functions:
            intersection_span_span, intersection_spanset_span, 
            intersection_period_timestamp
        """
        from .periodset import PeriodSet
        from .timestampset import TimestampSet
        if isinstance(other, datetime):
            result = intersection_period_timestamp(self._inner,
                datetime_to_timestamptz(other))
            return timestamptz_to_datetime(result) if result is not None else None
        elif isinstance(other, TimestampSet):
            return self.intersection(other.to_periodset())
        elif isinstance(other, Period):
            result = intersection_span_span(self._inner, other._inner)
            return Period(_inner=result) if result is not None else None
        elif isinstance(other, PeriodSet):
            result = intersection_spanset_span(other._inner, self._inner)
            return PeriodSet(_inner=result) if result is not None else None
        else:
            super().intersection(other)

    def minus(self, other: Time) -> PeriodSet:
        """
        Returns the temporal difference of ``self`` and ``other``.

        Args:
            other: temporal object to diff with

        Returns:
            A :class:`PeriodSet` instance.

        MEOS Functions:
            minus_period_timestamp, minus_span_spanset, minus_span_span
        """
        from .periodset import PeriodSet
        from .timestampset import TimestampSet
        if isinstance(other, datetime):
            result = minus_period_timestamp(self._inner,
                datetime_to_timestamptz(other))
        elif isinstance(other, TimestampSet):
            return self.minus(other.to_periodset())
        elif isinstance(other, Period):
            result = minus_span_span(self._inner, other._inner)
        elif isinstance(other, PeriodSet):
            result = minus_span_spanset(self._inner, other._inner)
        else:
            result = super().minus(other)
        return PeriodSet(_inner=result) if result is not None else None

    def union(self, other: Time) -> PeriodSet:
        """
        Returns the temporal union of ``self`` and ``other``.

        Args:
            other: temporal object to merge with

        Returns:
            A :class:`PeriodSet` instance.

        MEOS Functions:
            union_period_timestamp, union_spanset_span, union_span_span
        """
        from .periodset import PeriodSet
        from .timestampset import TimestampSet
        if isinstance(other, datetime):
            result = union_period_timestamp(self._inner,
                datetime_to_timestamptz(other))
        elif isinstance(other, TimestampSet):
            result = super().union(other)
        elif isinstance(other, Period):
            result = super().union(other)
        elif isinstance(other, PeriodSet):
            result = super().union(other)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return PeriodSet(_inner=result) if result is not None else None

    # ------------------------- Plot Operations -------------------------------
    def plot(self, *args, **kwargs):
        from ...plotters import TimePlotter
        return TimePlotter.plot_period(self, *args, **kwargs)



from __future__ import annotations

from abc import ABC
from datetime import datetime
from typing import Optional, Union, Any, TYPE_CHECKING, TypeVar, List

from pymeos_cffi import *

from .temporal import Temporal

TBase = TypeVar('TBase')
TG = TypeVar('TG', bound='Temporal[Any]')
TI = TypeVar('TI', bound='TInstant[Any]')
TS = TypeVar('TS', bound='TSequence[Any]')
TSS = TypeVar('TSS', bound='TSequenceSet[Any]')
Self = TypeVar('Self', bound='TInstant[Any]')


class TInstant(Temporal[TBase, TG, TI, TS, TSS], ABC):
    """
    Base class for temporal instant types, i.e. temporal values that are
    defined at a single point in time.
    """
    __slots__ = ['_inner']

    _make_function = None
    _cast_function = None

    def __init__(self, string: Optional[str] = None, *,
                 value: Optional[Union[str, TBase]] = None,
                 timestamp: Optional[Union[str, datetime]] = None,
                 _inner=None):
        assert (_inner is not None) or ((string is not None) != \
            (value is not None and timestamp is not None)), \
            "Either string must be not None or both point and timestamp must be not"
        if _inner is not None:
            self._inner = as_tinstant(_inner)
        elif string is not None:
            self._inner = as_tinstant(self.__class__._parse_function(string))
        else:
            ts = datetime_to_timestamptz(timestamp) if isinstance(timestamp, datetime) \
                else pg_timestamptz_in(timestamp, -1)
            self._inner = self.__class__._make_function(self.__class__._cast_function(value), ts)

    def value(self) -> TBase:
        """
        Returns the value of the temporal instant.

        Returns:
            The value of the temporal instant.
        """
        return self.start_value()

    def timestamp(self) -> datetime:
        """
        Returns the timestamp of the temporal instant.

        Returns:
            A :class:`~datetime.datetime` object.

        MEOS Functions:
            temporal_timestamps
        """
        ts, count = temporal_timestamps(self._inner)
        assert count == 1
        return timestamptz_to_datetime(ts[0])

    def start_instant(self: Self) -> Self:
        return self

    def end_instant(self: Self) -> Self:
        return self

    def instant_n(self: Self, n: int) -> Self:
        if n == 0:
            return self
        else:
            raise Exception("ERROR: Out of range")

    def instants(self: Self) -> List[Self]:
        return [self]

    def start_timestamp(self) -> datetime:
        return self.timestamp()

    def end_timestamp(self) -> datetime:
        return self.timestamp()

    def timestamp_n(self, n) -> datetime:
        if n == 0:
            return self.timestamp()
        else:
            raise Exception("ERROR: Out of range")

    def timestamps(self) -> List[datetime]:
        return [self.timestamp()]



from __future__ import annotations

from abc import ABC
from typing import Union, TYPE_CHECKING, TypeVar

from pymeos_cffi import *

from ..collections import IntSet, FloatSet
from ..temporal import Temporal

if TYPE_CHECKING:
    from ..boxes import Box, TBox
    from ..collections import Time, IntSpan, IntSpanSet, FloatSpan, FloatSpanSet
    from .tint import TInt
    from .tfloat import TFloat

TBase = TypeVar('TBase', int, float)
TG = TypeVar('TG', 'TNumber[int]', 'TNumber[float]')
TI = TypeVar('TI', 'TInstant[int]', 'TInstant[float]')
TS = TypeVar('TS', 'TSequence[int]', 'TSequence[float]')
TSS = TypeVar('TSS', 'TSequenceSet[int]', 'TSequenceSet[float]')
Self = TypeVar('Self', bound='TNumber[Any]')


class TNumber(Temporal[TBase, TG, TI, TS, TSS], ABC):

    # ------------------------- Accessors -------------------------------------
    def bounding_box(self) -> TBox:
        """
        Returns the bounding box of `self`.

        Returns:
            The bounding box of `self`.

        MEOS Functions:
            tbox_tnumber
        """
        from ..boxes import TBox
        return TBox(_inner=tnumber_to_tbox(self._inner))

    def integral(self) -> float:
        """
        Returns the integral of `self`.

        Returns:
            The integral of `self`.

        MEOS Function:
            tnumber_integral
        """
        return tnumber_integral(self._inner)

    def time_weighted_average(self) -> float:
        """
        Returns the time weighted average of `self`.

        Returns:
            The time weighted average of `self`.

        MEOS Function:
            tnumber_twavg
        """
        return tnumber_twavg(self._inner)

    # ------------------------- Transformations -------------------------------
    def shift_value(self: Self, delta: Union[int, float]) -> Self:
        """
        Returns a new :class:`TNumber` with the value dimension shifted by
        ``delta``.

        Args:
            delta: value to shift

        MEOS Functions:
            tint_shift_value, tfloat_shift_value
        """
        from .tint import TInt
        from .tfloat import TFloat
        if isinstance(self, TInt):
            shifted = tint_shift_value(self._inner, int(delta))
        elif isinstance(self, TFloat):
            shifted = tfloat_shift_value(self._inner, float(delta))
        else:
            raise TypeError(f'Operation not supported with type {self.__class__}')
        return Temporal._factory(shifted)

    def scale_value(self: Self, width: Union[int, float]) -> Self:
        """
        Returns a new :class:`TNumber` scaled so the value dimension has
        width ``width``.

        Args:
            width: value representing the width of the new temporal number

        MEOS Functions:
            tint_scale_value, tfloat_scale_value
        """
        from .tint import TInt
        from .tfloat import TFloat
        if isinstance(self, TInt):
            scaled = tint_scale_value(self._inner, int(width))
        elif isinstance(self, TFloat):
            scaled = tfloat_scale_value(self._inner, float(width))
        else:
            raise TypeError(f'Operation not supported with type {self.__class__}')
        return Temporal._factory(scaled)

    def shift_scale_value(self: Self, shift: Union[int, float] = None,
                          width: Union[int, float] = None) -> Self:
        """
        Returns a new :class:`TNumber` with the value dimension shifted by
        ``shift`` and scaled so the value dimension has width ``width``.

        Args:
            shift: value to shift
            width: value representing the width of the new temporal number

        MEOS Functions:
            tint_shift_scale_value, tfloat_shift_scale_value
        """
        from .tint import TInt
        from .tfloat import TFloat
        assert shift is not None or width is not None, \
            'shift and width must not be both None'

        if isinstance(self, TInt):
            scaled = tint_shift_scale_value(self._inner,
                                            int(shift) if shift else None, int(width) if width else None)
        elif isinstance(self, TFloat):
            scaled = tfloat_shift_scale_value(self._inner,
                                              float(shift) if shift else None, float(width) if width else None)
        else:
            raise TypeError(f'Operation not supported with type {self.__class__}')
        return Temporal._factory(scaled)

    # ------------------------- Restrictions ----------------------------------
    def at(self, other: Union[IntSet, FloatSet, IntSpan, FloatSpan, IntSpanSet, FloatSpanSet,
    TBox, Time]) -> TG:
        """
        Returns a new temporal object with the values of `self` restricted to
        the value or time `other`.

        Args:
            other: A time or value object to restrict the values of `self` to.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            temporal_at_values, tnumber_at_span, tnumber_at_spanset, tnumber_at_tbox,
            temporal_at_timestamp, temporal_at_timestampset,
            temporal_at_period, temporal_at_periodset
        """
        from ..boxes import TBox
        from ..collections import IntSet, FloatSet, IntSpan, FloatSpan, IntSpanSet, FloatSpanSet
        if isinstance(other, IntSet) or isinstance(other, FloatSet):
            result = temporal_at_values(self._inner, other._inner)
        elif isinstance(other, IntSpan) or isinstance(other, FloatSpan):
            result = tnumber_at_span(self._inner, other._inner)
        elif isinstance(other, IntSpanSet) or isinstance(other, FloatSpanSet):
            result = tnumber_at_spanset(self._inner, other._inner)
        elif isinstance(other, TBox):
            result = tnumber_at_tbox(self._inner, other._inner)
        else:
            return super().at(other)
        return Temporal._factory(result)

    def minus(self, other: Union[IntSet, FloatSet, IntSpan, FloatSpan, IntSpanSet, FloatSpanSet,
    TBox, Time]) -> TG:
        """
        Returns a new temporal object with the values of `self` restricted to
        the complement of the value or time `other`.

        Args:
            other: A time or value object to restrict the values of `self` to
            the complement of.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            temporal_minus_values, tnumber_minus_span, tnumber_minus_spanset, tnumber_minus_tbox,
            temporal_minus_timestamp, temporal_minus_timestampset,
            temporal_minus_period, temporal_minus_periodset
        """
        from ..boxes import TBox
        from ..collections import IntSet, FloatSet, IntSpan, FloatSpan, IntSpanSet, FloatSpanSet
        if isinstance(other, IntSet) or isinstance(other, FloatSet):
            result = temporal_minus_values(self._inner, other._inner)
        elif isinstance(other, IntSpan) or isinstance(other, FloatSpan):
            result = tnumber_minus_span(self._inner, other._inner)
        elif isinstance(other, IntSpanSet) or isinstance(other, FloatSpanSet):
            result = tnumber_minus_spanset(self._inner, other._inner)
        elif isinstance(other, TBox):
            result = tnumber_minus_tbox(self._inner, other._inner)
        else:
            return super().minus(other)
        return Temporal._factory(result)

    # ------------------------- Position Operations ---------------------------
    def is_left(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is left to the bounding box
        of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if left, False otherwise.

        See Also:
            :meth:`Period.is_before`
        """
        return self.bounding_box().is_left(other)

    def is_over_or_left(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is over or left to the
        bounding box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if over or left, False otherwise.

        See Also:
            :meth:`Period.is_over_or_before`
        """
        return self.bounding_box().is_over_or_left(other)

    def is_right(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is right to the bounding
        box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if right, False otherwise.

        See Also:
            :meth:`Period.is_after`
        """
        return self.bounding_box().is_right(other)

    def is_over_or_right(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is over or right to the
        bounding box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if over or right, False otherwise.

        See Also:
            :meth:`Period.is_over_or_before`
        """
        return self.bounding_box().is_over_or_right(other)

    # ------------------------- Mathematical Operations -------------------------
    def add(self, other: Union[int, float, TNumber]) -> TNumber:
        """
        Returns a new temporal object with the values of `self` plus `other`.

        Args:
            other: A :class:`int`, :class:`float` or :class:`TNumber` to add
            to `self`.

        Returns:
            A new temporal object of the same subtype as `self`.
        """
        from .tint import TInt
        from .tfloat import TFloat
        if isinstance(self, TInt) and isinstance(other, int):
            result = add_tint_int(self._inner, other)
        elif isinstance(self, TFloat) and isinstance(other, (int, float)):
            result = add_tfloat_float(self._inner, float(other))
        elif isinstance(other, TNumber):
            result = add_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def radd(self, other: Union[int, float]) -> TNumber:
        """
        Returns a new temporal object with the values of `self` plus `other`.

        Args:
            other: A :class:`int` or :class:`float` to add to `self`.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            add_int_tint, add_float_tfloat
        """
        from .tint import TInt
        from .tfloat import TFloat
        if isinstance(self, TInt) and isinstance(other, int):
            result = add_int_tint(other, self._inner)
        elif isinstance(self, TFloat) and isinstance(other, (int, float)):
            result = add_float_tfloat(float(other), self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def sub(self, other: Union[int, float, TNumber]) -> TNumber:
        """
        Returns a new temporal object with the values of `self` minus `other`.

        Args:
            other: A :class:`int`, :class:`float` or :class:`TNumber` to
            subtract from `self`.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            sub_tint_int, sub_tfloat_float, sub_tnumber_tnumber
        """
        from .tint import TInt
        from .tfloat import TFloat
        if isinstance(self, TInt) and isinstance(other, int):
            result = sub_tint_int(self._inner, other)
        elif isinstance(self, TFloat) and isinstance(other, (int, float)):
            result = sub_tfloat_float(self._inner, float(other))
        elif isinstance(other, TNumber):
            result = sub_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def rsub(self, other: Union[int, float]) -> TNumber:
        """
        Returns a new temporal object with the values of `other` minus `self`.

        Args:
            other: A :class:`int` or :class:`float` to subtract `self` to.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            sub_int_tint, sub_float_tfloat
        """
        from .tint import TInt
        from .tfloat import TFloat
        if isinstance(self, TInt) and isinstance(other, int):
            result = sub_int_tint(other, self._inner)
        elif isinstance(self, TFloat) and isinstance(other, (int, float)):
            result = sub_float_tfloat(float(other), self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def mul(self, other: Union[int, float, TNumber]) -> TNumber:
        """
        Returns a new temporal object with the values of `self` multiplied by
        `other`.

        Args:
            other: A :class:`int`, :class:`float` or :class:`TNumber` to
            multiply `self` by.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            mult_tint_int, mult_tfloat_float, mult_tnumber_tnumber
        """
        from .tint import TInt
        from .tfloat import TFloat
        if isinstance(self, TInt) and isinstance(other, int):
            result = mult_tint_int(self._inner, other)
        elif isinstance(self, TFloat) and isinstance(other, (int, float)):
            result = mult_tfloat_float(self._inner, float(other))
        elif isinstance(other, TNumber):
            result = mult_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def rmul(self, other: Union[int, float]) -> TNumber:
        """
        Returns a new temporal object with the values of `self` multiplied by
        `other`.

        Args:
            other: A :class:`int` or :class:`float` to multiply by `self`.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            mult_int_tint, mult_float_tfloat
        """
        from .tint import TInt
        from .tfloat import TFloat
        if isinstance(self, TInt) and isinstance(other, int):
            result = mult_int_tint(other, self._inner)
        elif isinstance(self, TFloat) and isinstance(other, (int, float)):
            result = mult_float_tfloat(float(other), self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def div(self, other: Union[int, float, TNumber]) -> TNumber:
        """
        Returns a new temporal object with the values of `self` divided by
        `other`.

        Args:
            other: A :class:`int`, :class:`float` or :class:`TNumber` to divide
            `self` by.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            div_tint_int, div_tfloat_float, div_tnumber_tnumber
        """
        from .tint import TInt
        from .tfloat import TFloat
        if isinstance(self, TInt) and isinstance(other, int):
            result = div_tint_int(self._inner, other)
        elif isinstance(self, TFloat) and isinstance(other, (int, float)):
            result = div_tfloat_float(self._inner, float(other))
        elif isinstance(other, TNumber):
            result = div_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def rdiv(self, other: Union[int, float]) -> TNumber:
        """
        Returns a new temporal object with the values of `other` divided by
        `self`.

        Args:
            other: A :class:`int` or :class:`float` to divide by `self`.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            div_int_tint, div_float_tfloat
        """
        from .tint import TInt
        from .tfloat import TFloat
        if isinstance(self, TInt) and isinstance(other, int):
            result = div_int_tint(other, self._inner)
        elif isinstance(self, TFloat) and isinstance(other, (int, float)):
            result = div_float_tfloat(float(other), self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def __add__(self, other):
        """
        Returns a new temporal object with the values of `self` plus `other`.

        Args:
            other: A :class:`int`, :class:`float` or :class:`TNumber` to add to
            `self`.

        Returns:
            A new temporal object of the same subtype as `self`.
        """
        return self.add(other)

    def __radd__(self, other):
        """
        Returns a new temporal object with the values of `self` plus `other`.

        Args:
            other: A :class:`int` or :class:`float` to add to `self`.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            add_int_tint, add_float_tfloat
        """
        return self.radd(other)

    def __sub__(self, other):
        """
        Returns a new temporal object with the values of `self` minus `other`.

        Args:
            other: A :class:`int`, :class:`float` or :class:`TNumber` to
            subtract from `self`.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            sub_tint_int, sub_tfloat_float, sub_tnumber_tnumber
        """
        return self.sub(other)

    def __rsub__(self, other):
        """
        Returns a new temporal object with the values of `other` minus `self`.

        Args:
            other: A :class:`int` or :class:`float` to subtract `self` to.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            sub_int_tint, sub_float_tfloat
        """
        return self.rsub(other)

    def __mul__(self, other):
        """
        Returns a new temporal object with the values of `self` multiplied by
        `other`.

        Args:
            other: A :class:`int`, :class:`float` or :class:`TNumber` to
            multiply `self` by.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            mult_tint_int, mult_tfloat_float, mult_tnumber_tnumber
        """
        return self.mul(other)

    def __rmul__(self, other):
        """
        Returns a new temporal object with the values of `self` multiplied
        by `other`.

        Args:
            other: A :class:`int` or :class:`float` to multiply by `self`.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            mult_int_tint, mult_float_tfloat
        """
        return self.rmul(other)

    def __truediv__(self, other):
        """
        Returns a new temporal object with the values of `self` divided by
        `other`.

        Args:
            other: A :class:`int`, :class:`float` or :class:`TNumber` to divide
            `self` by.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            div_tint_int, div_tfloat_float, div_tnumber_tnumber
        """
        return self.div(other)

    def __rtruediv__(self, other):
        """
        Returns a new temporal object with the values of `other` divided by
        `self`.

        Args:
            other: A :class:`int` or :class:`float` to divide by `self`.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            div_int_tint, div_float_tfloat
        """
        return self.rdiv(other)

    def abs(self) -> TNumber:
        """
        Returns the absolute value of `self`.

        Returns:
            A new :class:`TNumber` instance.

        MEOS Functions:
            tnumber_abs
        """
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(tnumber_abs(self._inner))

    def delta_value(self) -> TNumber:
        """
        Returns the value difference between consecutive instants of `self`.

        Returns:
            A new :class:`TNumber` instance.

        MEOS Functions:
            tnumber_delta_value
        """
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(tnumber_delta_value(self._inner))

    # ------------------------- Distance Operations --------------------------
    def distance(self, other: Union[int, float, TNumber]) -> TFloat:
        """
        Returns the temporal distance between `self` and `other`.

        Args:
            other: A :class:`int`, :class:`float` or :class:`TNumber` to
            compare to `self`.

        Returns:
            A :class:`TFloat` with the distance between `self` and `other`.

        MEOS Functions:
            distance_tfloat_float, distance_tnumber_tnumber
        """
        if isinstance(other, int):
            result = distance_tfloat_float(self._inner, float(other))
        elif isinstance(other, float):
            result = distance_tfloat_float(self._inner, other)
        elif isinstance(other, TNumber):
            result = distance_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def nearest_approach_distance(self, other: Union[int, float,
    TNumber, TBox]) -> float:
        """
        Returns the nearest approach distance between `self` and `other`.

        Args:
            other: A :class:`int`, :class:`float`, :class:`TNumber` or
            :class:`TBox` to compare to `self`.

        Returns:
            A :class:`float` with the nearest approach distance between `self`
            and `other`.

        MEOS Functions:
            nad_tfloat_float, nad_tfloat_tfloat, nad_tnumber_tbox
        """
        if isinstance(other, int):
            return nad_tfloat_float(self._inner, float(other))
        elif isinstance(other, float):
            return nad_tfloat_float(self._inner, other)
        elif isinstance(other, TNumber):
            return nad_tfloat_tfloat(self._inner, other._inner)
        elif isinstance(other, TBox):
            return nad_tnumber_tbox(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')



from __future__ import annotations

from typing import Optional, Union, List, TYPE_CHECKING, get_args

import shapely.geometry.base as shp
from pymeos_cffi import *

from ..main import TPoint
from ..temporal import Temporal
from ..collections import *

if TYPE_CHECKING:
    from .box import Box


class STBox:
    """
    Class for representing a spatio-temporal box. Temporal bounds may be inclusive or
    exclusive.

    ``STBox`` objects can be created with a single argument of type string as in MobilityDB.

        >>> STBox('STBOX ZT(((1.0,2.0,3.0),(4.0,5.0,6.0)),[2001-01-01, 2001-01-02])')

    Another possibility is to provide the different dimensions with the corresponding parameters:
        - ``xmin``, ``xmax``, ``ymin``, ``ymax`` for spatial dimension
        - ``zmin``, ``zmax`` for the third spatial dimension
        - ``tmin``, ``tmax`` for temporal dimension
        - ``tmin_inc``, ``tmax_inc`` to specify if the temporal bounds are inclusive or exclusive
        - ``geodetic`` to specify if the spatial dimension is geodetic
        - ``srid`` to specify the spatial reference system identifier

    Note that at least the 2D spatial dimension or the temporal dimension must be provided.

        >>> STBox(xmin=1.0, xmax=4.0, ymin=2.0, ymax=5.0, tmin=datetime(2001, 1, 1), tmax=datetime(2001, 1, 2))

    """
    __slots__ = ['_inner']

    _mobilitydb_name = 'stbox'

    def _get_box(self, other: Union[shp.BaseGeometry, STBox, Temporal, Time],
                 allow_space_only: bool = True,
                 allow_time_only: bool = False) -> STBox:
        if allow_space_only and isinstance(other, shp.BaseGeometry):
            other_box = geo_to_stbox(geo_to_gserialized(other, self.geodetic()))
        elif isinstance(other, STBox):
            other_box = other._inner
        elif isinstance(other, TPoint):
            other_box = tpoint_to_stbox(other._inner)
        elif allow_time_only and isinstance(other, Temporal):
            other_box = period_to_stbox(temporal_to_period(other._inner))
        elif allow_time_only and isinstance(other, datetime):
            other_box = timestamp_to_stbox(datetime_to_timestamptz(other))
        elif allow_time_only and isinstance(other, TimestampSet):
            other_box = timestampset_to_stbox(other._inner)
        elif allow_time_only and isinstance(other, Period):
            other_box = period_to_stbox(other._inner)
        elif allow_time_only and isinstance(other, PeriodSet):
            other_box = periodset_to_stbox(other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return other_box

    # ------------------------- Constructors ----------------------------------
    def __init__(self, string: Optional[str] = None, *,
                 xmin: Optional[Union[str, float]] = None,
                 xmax: Optional[Union[str, float]] = None,
                 ymin: Optional[Union[str, float]] = None,
                 ymax: Optional[Union[str, float]] = None,
                 zmin: Optional[Union[str, float]] = None,
                 zmax: Optional[Union[str, float]] = None,
                 tmin: Optional[Union[str, datetime]] = None,
                 tmax: Optional[Union[str, datetime]] = None,
                 tmin_inc: bool = True, tmax_inc: bool = True,
                 geodetic: bool = False, srid: Optional[int] = None,
                 _inner=None):

        assert (_inner is not None) or (string is not None) != (
                (xmin is not None and xmax is not None and ymin is not None and
                 ymax is not None) or
                (tmin is not None and tmax is not None)), \
            "Either string must be not None or at least a bound pair (xmin/max" \
            " and ymin/max, or tmin/max) must be not None"

        if _inner is not None:
            self._inner = _inner
        elif string is not None:
            self._inner = stbox_in(string)
        else:
            period = None
            hast = tmin is not None and tmax is not None
            hasx = xmin is not None and xmax is not None and ymin is not None \
            and ymax is not None
            hasz = zmin is not None and zmax is not None
            if hast:
                period = Period(lower=tmin, upper=tmax, lower_inc=tmin_inc,
                upper_inc=tmax_inc)._inner
            self._inner = stbox_make(hasx, hasz, geodetic, srid or 0, 
                float(xmin or 0), float(xmax or 0),
                float(ymin or 0), float(ymax or 0),
                float(zmin or 0), float(zmax or 0), period)

    def __copy__(self) -> STBox:
        """
        Returns a copy of ``self``.

        Returns:
            A :class:`STBox` instance.

        MEOS Functions:
            stbox_copy
        """
        inner_copy = stbox_copy(self._inner)
        return STBox(_inner=inner_copy)

    @staticmethod
    def from_wkb(wkb: bytes) -> STBox:
        """
        Returns a `STBox` from its WKB representation.

        Args:
            wkb: WKB representation

        Returns:
            A new :class:`STBox` instance

        MEOS Functions:
            stbox_from_wkb
        """
        result = stbox_from_wkb(wkb)
        return STBox(_inner=result)

    @staticmethod
    def from_hexwkb(hexwkb: str) -> STBox:
        """
        Returns a `STBox` from its WKB representation in hex-encoded ASCII.

        Args:
            hexwkb: WKB representation in hex-encoded ASCII

        Returns:
            A new :class:`STBox` instance

        MEOS Functions:
            stbox_from_hexwkb
        """
        result = stbox_from_hexwkb(hexwkb)
        return STBox(_inner=result)

    @staticmethod
    def from_geometry(geom: shp.BaseGeometry, geodetic: bool = False) -> STBox:
        """
        Returns a `STBox` from a `shp.BaseGeometry`.

        Args:
            geom: A `shp.BaseGeometry` instance.
            geodetic: Whether to create a geodetic or geometric `STBox`.

        Returns:
            A new :class:`STBox` instance.

        MEOS Functions:
            pgis_geometry_in, geo_to_stbox
        """
        gs = geo_to_gserialized(geom, geodetic)
        return STBox(_inner=geo_to_stbox(gs))

    @staticmethod
    def from_time(time: Time) -> STBox:
        """
        Returns a `STBox` from a `Time` instance.

        Args:
            time: A `Time` instance.

        Returns:
            A new :class:`STBox` instance.

        MEOS Functions:
            timestamp_to_stbox, timestampset_to_stbox, period_to_stbox,
            periodset_to_stbox
        """
        if isinstance(time, datetime):
            result = timestamp_to_stbox(datetime_to_timestamptz(time))
        elif isinstance(time, TimestampSet):
            result = timestampset_to_stbox(time._inner)
        elif isinstance(time, Period):
            result = period_to_stbox(time._inner)
        elif isinstance(time, PeriodSet):
            result = periodset_to_stbox(time._inner)
        else:
            raise TypeError(f'Operation not supported with type {time.__class__}')
        return STBox(_inner=result)

    @staticmethod
    def from_geometry_time(geometry: shp.BaseGeometry, time: Union[datetime, Period],
                           geodetic: bool = False) -> STBox:
        """
        Returns a `STBox` from a space and time dimension.

        Args:
            geometry: A `shp.BaseGeometry` instance representing the space dimension.
            time: A `Time` instance representing the time dimension.
            geodetic: Whether to create a geodetic or geometric `STBox`.

        Returns:
            A new :class:`STBox` instance.

        MEOS Functions:
            geo_timestamp_to_stbox, geo_period_to_stbox
        """
        gs = geo_to_gserialized(geometry, geodetic)
        if isinstance(time, datetime):
            result = geo_timestamp_to_stbox(gs, datetime_to_timestamptz(time))
        elif isinstance(time, Period):
            result = geo_period_to_stbox(gs, time._inner)
        else:
            raise TypeError(f'Operation not supported with types ' \
                '{geometry.__class__} and {time.__class__}')
        return STBox(_inner=result)

    @staticmethod
    def from_tpoint(temporal: TPoint) -> STBox:
        """
        Returns the bounding box of a `TPoint` instance as an `STBox`.

        Args:
            temporal: A `TPoint` instance.

        Returns:
            A new :class:`STBox` instance.

        MEOS Functions:
            tpoint_to_stbox
        """
        return STBox(_inner=tpoint_to_stbox(temporal._inner))

    @staticmethod
    def from_expanding_bounding_box(value: Union[shp.BaseGeometry, TPoint, STBox],
        expansion: float, geodetic: Optional[bool] = False) -> STBox:
        """
        Returns a `STBox` from a `shp.BaseGeometry`, `TPoint` or `STBox` instance,
        expanding its bounding box by the given amount.

        Args:
            value: A `shp.BaseGeometry`, `TPoint` or `STBox` instance.
            expansion: The amount to expand the bounding box.
            geodetic: Whether to create a geodetic or geometric `STBox`.
            Only used when value is a `shp.BaseGeometry` instance.

        Returns:
            A new :class:`STBox` instance.

        MEOS Functions:
            geo_expand_space, tpoint_expand_space, stbox_expand_space
        """
        if isinstance(value, shp.BaseGeometry):
            gs = geo_to_gserialized(value, geodetic)
            result = geo_expand_space(gs, expansion)
        elif isinstance(value, TPoint):
            result = tpoint_expand_space(value._inner, expansion)
        elif isinstance(value, STBox):
            result = stbox_expand_space(value._inner, expansion)
        else:
            raise TypeError(f'Operation not supported with type {value.__class__}')
        return STBox(_inner=result)

    # ------------------------- Output ----------------------------------------
    def __str__(self, max_decimals: int = 15):
        """
        Returns a string representation of ``self``.

        Returns:
            A :class:`str` instance.

        MEOS Functions:
            stbox_out
        """
        return stbox_out(self._inner, max_decimals)

    def __repr__(self):
        """
        Returns a string representation of ``self``.

        Returns:
            A :class:`str` instance.

        MEOS Functions:
            stbox_out
        """
        return (f'{self.__class__.__name__}'
                f'({self})')

    def as_wkb(self) -> bytes:
        """
        Returns the WKB representation of ``self``.

        Returns:
            A :class:`str` object with the WKB representation of ``self``.

        MEOS Functions:
            stbox_as_wkb
        """
        return stbox_as_wkb(self._inner, 4)

    def as_hexwkb(self) -> str:
        """
        Returns the WKB representation of ``self`` in hex-encoded ASCII.

        Returns:
            A :class:`str` object with the WKB representation of ``self`` in
            hex-encoded ASCII.

        MEOS Functions:
            stbox_as_hexwkb
        """
        return stbox_as_hexwkb(self._inner, -1)[0]

    # ------------------------- Conversions ----------------------------------
    def to_geometry(self, precision: int = 15) -> shp.BaseGeometry:
        """
        Returns the spatial dimension of ``self`` as a `shapely`
        :class:`~shapely.BaseGeometry` instance.

        Args:
            precision: The precision of the geometry coordinates.

        Returns:
            A new :class:`~shapely.BaseGeometry` instance.

        MEOS Functions:
            stbox_to_geo
        """
        return gserialized_to_shapely_geometry(stbox_to_geo(self._inner),
            precision)

    def to_period(self) -> Period:
        """
        Returns the temporal dimension of ``self`` as a `Period` instance.

        Returns:
            A new :class:`Period` instance.

        MEOS Functions:
            stbox_to_period
        """
        return Period(_inner=stbox_to_period(self._inner))

    # ------------------------- Accessors -------------------------------------
    def has_xy(self) -> bool:

        """
        Returns whether ``self`` has a spatial (XY) dimension.

        Returns:
            True if ``self`` has a spatial dimension, False otherwise.

        MEOS Functions:
            stbox_hasx
        """
        return stbox_hasx(self._inner)

    def has_z(self) -> bool:
        """
        Returns whether ``self`` has a Z dimension.

        Returns:
            True if ``self`` has a Z dimension, False otherwise.

        MEOS Functions:
            stbox_hasz
        """
        return stbox_hasz(self._inner)

    def has_t(self) -> bool:
        """
        Returns whether ``self`` has a time dimension.

        Returns:
            True if ``self`` has a time dimension, False otherwise.

        MEOS Functions:
            stbox_hast
        """
        return stbox_hast(self._inner)

    def geodetic(self) -> bool:
        """
        Returns whether ``self`` is geodetic.

        Returns:
            True if ``self`` is geodetic, False otherwise.

        MEOS Functions:
            stbox_isgeodetic
        """
        return stbox_isgeodetic(self._inner)

    def xmin(self) -> Optional[float]:
        """
        Returns the minimum X coordinate of ``self``.

        Returns:
            A :class:`float` with the minimum X coordinate of ``self``.

        MEOS Functions:
            stbox_xmin
        """
        return stbox_xmin(self._inner)

    def ymin(self) -> Optional[float]:
        """
        Returns the minimum Y coordinate of ``self``.

        Returns:
            A :class:`float` with the minimum Y coordinate of ``self``.

        MEOS Functions:
            stbox_ymin
        """
        return stbox_ymin(self._inner)

    def zmin(self) -> Optional[float]:
        """
        Returns the minimum Z coordinate of ``self``.

        Returns:
            A :class:`float` with the minimum Z coordinate of ``self``.

        MEOS Functions:
            stbox_zmin
        """
        return stbox_zmin(self._inner)

    def tmin(self) -> Optional[datetime]:
        """
        Returns the starting time of ``self``.

        Returns:
            A :class:`datetime` with the minimum time coordinate of ``self``.

        MEOS Functions:
            stbox_tmin
        """
        result = stbox_tmin(self._inner)
        return timestamptz_to_datetime(result) if result else None

    def tmin_inc(self) -> bool:
        """
        Returns whether starting time of ``self`` is inclusive or not

        Returns:
            True if the starting time of ``self`` is inclusive and False
            otherwise

        MEOS Functions:
            stbox_tmin_inc
        """
        return stbox_tmin_inc(self._inner)

    def xmax(self) -> Optional[float]:
        """
        Returns the maximum X coordinate of ``self``.

        Returns:
            A :class:`float` with the maximum X coordinate of ``self``.

        MEOS Functions:
            stbox_xmax
        """
        return stbox_xmax(self._inner)

    def ymax(self) -> Optional[float]:
        """
        Returns the maximum Y coordinate of ``self``.

        Returns:
            A :class:`float` with the maximum Y coordinate of ``self``.

        MEOS Functions:
            stbox_ymax
        """
        return stbox_ymax(self._inner)

    def zmax(self) -> Optional[float]:
        """
        Returns the maximum Z coordinate of ``self``.

        Returns:
            A :class:`float` with the maximum Z coordinate of ``self``.

        MEOS Functions:
            stbox_zmax
        """
        return stbox_zmax(self._inner)

    def tmax(self) -> Optional[datetime]:
        """
        Returns the ending time of ``self``.

        Returns:
            A :class:`datetime` with the maximum time coordinate of ``self``.

        MEOS Functions:
            stbox_tmax
        """
        result = stbox_tmax(self._inner)
        return timestamptz_to_datetime(result) if result else None

    def tmax_inc(self) -> bool:
        """
        Returns whether ending time of ``self`` is inclusive or not

        Returns:
            True if the ending time of ``self`` is inclusive and False otherwise

        MEOS Functions:
            stbox_tmax_inc
        """
        return stbox_tmax_inc(self._inner)

    # ------------------------- Spatial Reference System ----------------------
    def srid(self) -> int:
        """
        Returns the SRID of ``self``.

        Returns:
            An :class:`int` with the SRID of ``self``.

        MEOS Functions:
            stbox_srid
        """
        return self._inner.srid

    def set_srid(self, value: int) -> STBox:
        """
        Returns a copy of ``self`` with the SRID set to ``value``.

        Args:
            value: The new SRID.

        Returns:
            A new :class:`STBox` instance.

        MEOS Functions:
            stbox_set_srid
        """
        return STBox(_inner=stbox_set_srid(self._inner, value))

    # ------------------------- Transformations -------------------------------
    def get_space(self) -> STBox:
        """
        Get the spatial dimension of ``self``, removing the temporal dimension
        if any

        Returns:
            A new :class:`STBox` instance.

        MEOS Functions:
            stbox_get_space
        """
        result = stbox_get_space(self._inner)
        return STBox(_inner=result)

    def expand(self, other: Union[int, float, timedelta]) -> STBox:
        """
        Expands ``self`` with `other`.
        If `other` is a :class:`int` or a :class:`float`, the result is equal
        to ``self`` but with the spatial dimensions expanded by `other` in all
        directions. If `other` is a :class:`timedelta`, the result is equal to
        ``self``  but with the temporal dimension expanded by `other` in both
        directions.

        Args:
            other: The object to expand ``self`` with.

        Returns:
            A new :class:`STBox` instance.

        MEOS Functions:
            stbox_expand_space, stbox_expand_time
        """
        if isinstance(other, int) or isinstance(other, float):
            result = stbox_expand_space(self._inner, float(other))
        elif isinstance(other, timedelta):
            result = stbox_expand_time(self._inner,
                timedelta_to_interval(other))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return STBox(_inner=result)

    def shift_time(self, delta: timedelta) -> STBox:
        """
        Returns a new `STBox` with the time dimension shifted by `delta`.

        Args:
            delta: :class:`datetime.timedelta` instance to shift

        Returns:
            A new :class:`STBox` instance

        MEOS Functions:
            stbox_shift_scale_time

        See Also:
            :meth:`STBox.shift`
        """
        return self.shift_scale_time(shift=delta)

    def scale_time(self, duration: timedelta) -> STBox:
        """
        Returns a new `STBox` with the time dimension having duration 
        `duration`.

        Args:
            duration: :class:`datetime.timedelta` instance with new duration

        Returns:
            A new :class:`STBox` instance

        MEOS Functions:
            period_shift_scale

        See Also:
            :meth:`STBox.scale`
        """
        return self.shift_scale_time(duration=duration)

    def shift_scale_time(self, shift: Optional[timedelta] = None,
        duration: Optional[timedelta] = None) -> STBox:
        """
        Returns a new `STBox` with the time dimension shifted by `shift` and
        with duration `duration`.

        Args:
            shift: :class:`datetime.timedelta` instance to shift
            duration: :class:`datetime.timedelta` instance with new duration

        Returns:
            A new :class:`STBox` instance

        MEOS Functions:
            stbox_shift_scale_time

        See Also:
            :meth:`Period.shift_scale`
        """
        assert shift is not None or duration is not None, \
            'shift and scale deltas must not be both None'
        result = stbox_shift_scale_time(
            self._inner,
            timedelta_to_interval(shift) if shift else None,
            timedelta_to_interval(duration) if duration else None
        )
        return STBox(_inner=result)

    def round(self, maxdd : Optional[int] = 0) -> STBox:
        """
        Returns `self` rounded to the given number of decimal digits.

        Args:
            maxdd: Maximum number of decimal digits.

        Returns:
            A new :class:`STBox` instance

        MEOS Functions:
            stbox_round
        """
        new_inner = stbox_copy(self._inner)
        stbox_round(new_inner, maxdd)
        return STBox(_inner=new_inner)

    # ------------------------- Set Operations --------------------------------
    def union(self, other: STBox, strict: Optional[bool] = False) -> STBox:
        """
        Returns the union of `self` with `other`.

        Args:
            other: spatiotemporal box to merge with
            strict: Whether to fail if the boxes do not intersect.

        Returns:
            A :class:`STBox` instance.

        MEOS Functions:
            union_stbox_stbox
        """
        return STBox(_inner=union_stbox_stbox(self._inner, other._inner, strict))

    def __add__(self, other):
        """
        Returns the union of `self` with `other`. Fails if the union is not
        contiguous.

        Args:
            other: spatiotemporal box to merge with

        Returns:
            A :class:`STBox` instance.

        MEOS Functions:
            union_stbox_stbox
        """
        return self.union(other, False)

    # TODO: Check returning None for empty intersection is the desired behaviour
    def intersection(self, other: STBox) -> Optional[STBox]:
        """
        Returns the intersection of `self` with `other`.

        Args:
            other: temporal object to merge with

        Returns:
            A :class:`STBox` instance if the instersection is not empty, `None`
            otherwise.

        MEOS Functions:
            intersection_stbox_stbox
        """
        result = intersection_stbox_stbox(self._inner, other._inner)
        return STBox(_inner=result) if result else None

    def __mul__(self, other):
        """
        Returns the intersection of `self` with `other`.

        Args:
            other: temporal object to merge with

        Returns:
            A :class:`STBox` instance if the instersection is not empty, `None`
            otherwise.

        MEOS Functions:
            intersection_stbox_stbox
        """
        return self.intersection(other)

    # ------------------------- Topological Operations ------------------------
    def is_adjacent(self, other: Union[shp.BaseGeometry, STBox, Temporal, Time]) -> bool:
        """
        Returns whether ``self`` and `other` are adjacent. Two spatiotemporal
        boxes are adjacent if they share n dimensions and the intersection is
        of at most n-1 dimensions. Note that for `TPoint` instances, the
        bounding box of the temporal point is used.

        Args:
            other: The other spatiotemporal object to check adjacency with
            ``self``.

        Returns:
            ``True`` if ``self`` and `other` are adjacent, ``False`` otherwise.

        MEOS Functions:
            adjacent_stbox_stbox
        """
        return adjacent_stbox_stbox(self._inner, self._get_box(other,
            allow_time_only=True))

    def is_contained_in(self,
            container: Union[shp.BaseGeometry, STBox, Temporal, Time]) -> bool:
        """
        Returns whether ``self`` is contained in `container`. Note that for
        `TPoint` instances, the bounding box of the temporal point is used.

        Args:
            container: The spatiotemporal object to check containment with
            ``self``.

        Returns:
            ``True`` if ``self`` is contained in `container`, ``False`` otherwise.

        MEOS Functions:
            contained_stbox_stbox
        """
        return contained_stbox_stbox(self._inner, self._get_box(container,
            allow_time_only=True))

    def contains(self, content: Union[shp.BaseGeometry, STBox, Temporal, Time]) -> bool:
        """
        Returns whether ``self`` contains `content`. Note that for `TPoint`
        instances, the bounding box of the temporal point is used.

        Args:
            content: The spatiotemporal object to check containment with
            ``self``.

        Returns:
            ``True`` if ``self`` contains `content`, ``False`` otherwise.

        MEOS Functions:
            contains_stbox_stbox
        """
        return contains_stbox_stbox(self._inner, self._get_box(content,
            allow_time_only=True))

    def __contains__(self, item):
        """
        Returns whether ``self`` contains `item`.

        Args:
            item: The spatiotemporal object to check if it is contained in
            ``self``.

        Returns:
            ``True`` if ``self`` contains ``item``, ``False`` otherwise.

        MEOS Functions:
            contains_stbox_stbox

        See Also:
            :meth:`STBox.contains`
        """
        return self.contains(item)

    def overlaps(self, other: Union[shp.BaseGeometry, STBox, Temporal, Time]) -> bool:
        """
        Returns whether ``self`` overlaps `other`. Note that for `TPoint`
        instances, the bounding box of the temporal point is used.

        Args:
            other: The spatiotemporal object to check overlap with ``self``.

        Returns:
            ``True`` if ``self`` overlaps `other`, ``False`` otherwise.

        MEOS Functions:
            overlaps_stbox_stbox
        """
        return overlaps_stbox_stbox(self._inner, self._get_box(other,
            allow_time_only=True))

    def is_same(self, other: Union[shp.BaseGeometry, STBox, Temporal, Time]) -> bool:
        """
        Returns whether ``self`` is the same as `other`. Note that for `TPoint`
        instances, the bounding box of the temporal point is used.

        Args:
            other: The spatiotemporal object to check equality with ``self``.

        Returns:
            ``True`` if ``self`` is the same as `other`, ``False`` otherwise.

        MEOS Functions:
            same_stbox_stbox
        """
        return same_stbox_stbox(self._inner, self._get_box(other,
            allow_time_only=True))

    # ------------------------- Position Operations ---------------------------
    def is_left(self, other: Union[shp.BaseGeometry, STBox, TPoint]) -> bool:
        """
        Returns whether ``self`` is strictly to the left  of `other`. Checks
        the X dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is strictly to the left of `other`, 
            ``False`` otherwise.

        MEOS Functions:
            left_stbox_stbox
        """
        return left_stbox_stbox(self._inner, self._get_box(other))

    def is_over_or_left(self, other: Union[shp.BaseGeometry, STBox, TPoint]) -> bool:
        """
        Returns whether ``self`` is to the left `other` allowing for overlap.
        That is, ``self`` does not extend to the right of `other`. Checks the
        X dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is to the left of `other` allowing for
            overlap, ``False`` otherwise.

        MEOS Functions:
            overleft_stbox_stbox, tpoint_to_stbox
        """
        return overleft_stbox_stbox(self._inner, self._get_box(other))

    def is_right(self, other: Union[shp.BaseGeometry, STBox, TPoint]) -> bool:
        """
        Returns whether ``self`` is strictly to the right of `other`.
        Checks the X dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is strictly to the right of `other`,
            ``False`` otherwise.

        MEOS Functions:
            right_stbox_stbox
        """
        return right_stbox_stbox(self._inner, self._get_box(other))

    def is_over_or_right(self, other: Union[shp.BaseGeometry, STBox, TPoint]) -> bool:
        """
        Returns whether ``self`` is to the right of `other` allowing for
        overlap. That is, ``self`` does not extend to the left of `other`.
        Checks the X dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is to the right of `other` allowing for
            overlap, ``False`` otherwise.

        MEOS Functions:
            overright_stbox_stbox
        """
        return overright_stbox_stbox(self._inner, self._get_box(other))

    def is_below(self, other: Union[shp.BaseGeometry, STBox, TPoint]) -> bool:
        """
        Returns whether ``self`` is strictly below `other`.
        Checks the Y dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is strictly below `other`, ``False``
            otherwise.

        MEOS Functions:
            below_stbox_stbox
        """
        return below_stbox_stbox(self._inner, self._get_box(other))

    def is_over_or_below(self, other: Union[shp.BaseGeometry, STBox, TPoint]) -> bool:
        """
        Returns whether ``self`` is below `other` allowing for overlap.
        That is, ``self`` does not extend above `other`. Checks the Y dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is below `other` allowing for overlap,
            ``False`` otherwise.

        MEOS Functions:
            overbelow_stbox_stbox
        """
        return overbelow_stbox_stbox(self._inner, self._get_box(other))

    def is_above(self, other: Union[shp.BaseGeometry, STBox, TPoint]) -> bool:
        """
        Returns whether ``self`` is strictly above `other`.
        Checks the Y dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is strictly above `other`, ``False``
            otherwise.

        MEOS Functions:
            above_stbox_stbox
        """
        return above_stbox_stbox(self._inner, self._get_box(other))

    def is_over_or_above(self, other: Union[shp.BaseGeometry, STBox, TPoint]) -> bool:
        """
        Returns whether ``self`` is above `other` allowing for overlap.
        That is, ``self`` does not extend below `other`.
        Checks the Y dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is above `other` allowing for overlap,
            ``False`` otherwise.

        MEOS Functions:
            overabove_stbox_stbox
        """
        return overabove_stbox_stbox(self._inner, self._get_box(other))

    def is_front(self, other: Union[shp.BaseGeometry, STBox, TPoint]) -> bool:
        """
        Returns whether ``self`` is strictly in front of `other`.
        Checks the Z dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is strictly in front of `other`, ``False``
            otherwise.

        MEOS Functions:
            front_stbox_stbox
        """
        return front_stbox_stbox(self._inner, self._get_box(other))

    def is_over_or_front(self, other: Union[shp.BaseGeometry, STBox, TPoint]) -> bool:
        """
        Returns whether ``self`` is in front of `other` allowing for overlap.
        That is, ``self`` does not extend behind `other`.
        Checks the Z dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is in front of `other` allowing for overlap,
            ``False`` otherwise.

        MEOS Functions:
            overfront_stbox_stbox
        """
        return overfront_stbox_stbox(self._inner, self._get_box(other))

    def is_behind(self, other: Union[shp.BaseGeometry, STBox, TPoint]) -> bool:
        """
        Returns whether ``self`` is strictly behind `other`.
        Checks the Z dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is strictly behind `other`, ``False``
            otherwise.

        MEOS Functions:
            back_stbox_stbox
        """
        return back_stbox_stbox(self._inner, self._get_box(other))

    def is_over_or_behind(self, other: Union[shp.BaseGeometry, STBox, TPoint]) -> bool:
        """
        Returns whether ``self`` is behind `other` allowing for overlap.
        That is, ``self`` does not extend in front of `other`.
        Checks the Z dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is behind `other` allowing for overlap,
            ``False`` otherwise.

        MEOS Functions:
            overback_stbox_stbox
        """
        return overback_stbox_stbox(self._inner, self._get_box(other))

    def is_before(self, other: Union[Box, Temporal, Time]) -> bool:
        """
        Returns whether ``self`` is strictly before `other`.
        Checks the time dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is strictly before `other`, ``False``
            otherwise.
        """
        return self.to_period().is_before(other)

    def is_over_or_before(self, other: Union[Box, Temporal, Time]) -> bool:
        """
        Returns whether ``self`` is before `other` allowing for overlap.
        That is, ``self`` does not extend after `other`.
        Checks the time dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is before `other` allowing for overlap,
            ``False`` otherwise.
        """
        return self.to_period().is_over_or_before(other)

    def is_after(self, other: Union[Box, Temporal, Time]) -> bool:
        """
        Returns whether ``self`` is strictly after `other`.
        Checks the time dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is strictly after `other`, ``False``
            otherwise.
        """
        return self.to_period().is_after(other)

    def is_over_or_after(self, other: Union[Box, Temporal, Time]) -> bool:
        """
        Returns whether ``self`` is after `other` allowing for overlap.
        That is, ``self`` does not extend before `other`.
        Checks the time dimension.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is after `other` allowing for overlap, 
            ``False`` otherwise.
        """
        return self.to_period().is_over_or_after(other)

    # ------------------------- Distance Operations ---------------------------
    def nearest_approach_distance(self, other: Union[shp.BaseGeometry, STBox, TPoint]) \
        -> float:
        """
        Returns the distance between the nearest points of ``self`` and `other`.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            A :class:`float` with the distance between the nearest points of
            ``self`` and ``other``.

        MEOS Functions:
            nad_stbox_geo, nad_stbox_stbox
        """
        if isinstance(other, shp.BaseGeometry):
            gs = geo_to_gserialized(other, self.geodetic())
            return nad_stbox_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return nad_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return nad_tpoint_stbox(other._inner, self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    # ------------------------- Splitting --------------------------------------
    def quad_split_flat(self) -> List[STBox]:
        """
        Returns a list of 4 (or 8 if `self`has Z dimension) :class:`STBox`
        instances resulting from the quad  split of ``self``.

        Indices of returned array are as follows (back only present if Z
        dimension is present):

           >>> #    (front)          (back)
           >>> # -------------   -------------
           >>> # |  2  |  3  |   |  6  |  7  |
           >>> # ------------- + -------------
           >>> # |  0  |  1  |   |  4  |  5  |
           >>> # -------------   -------------

        Returns:
            A :class:`list` of :class:`STBox` instances.

        MEOS Functions:
            stbox_quad_split
        """
        boxes, count = stbox_quad_split(self._inner)
        return [STBox(_inner=boxes + i) for i in range(count)]

    def quad_split(self) -> Union[List[List[STBox]], List[List[List[STBox]]]]:
        """
        Returns a 2D (YxX) or 3D (ZxYxX) list of :class:`STBox` instances
        resulting from the quad split of ``self``.

        Indices of returned array are as follows:

           >>> #       (front)
           >>> # -------------------
           >>> # | [1][0] | [1][1] |
           >>> # -------------------
           >>> # | [0][0] | [0][1] |
           >>> # -------------------

        If Z dimension is present:

           >>> #          (front)                      (back)
           >>> # -------------------------   -------------------------
           >>> # | [0][1][0] | [0][1][1] |   | [1][1][0] | [1][1][1] |
           >>> # ------------------------- + -------------------------
           >>> # | [0][0][0] | [0][0][1] |   | [1][0][0] | [1][0][1] |
           >>> # -------------------------   -------------------------


        Returns:
            A 2D or 3D :class:`list` of :class:`STBox` instances.

        MEOS Functions:
            stbox_quad_split

        """
        boxes, count = stbox_quad_split(self._inner)
        if self.has_z():
            return [
                [[STBox(_inner=boxes + i) for i in range(2)],
                    [STBox(_inner=boxes + i) for i in range(2, 4)]],
                [[STBox(_inner=boxes + i) for i in range(4, 6)],
                    [STBox(_inner=boxes + i) for i in range(6, 8)]]
            ]
        else:
            return [[STBox(_inner=boxes + i) for i in range(2)],
                [STBox(_inner=boxes + i) for i in range(2, 4)]]

    def tile(self, size: Optional[float] = None, 
             duration: Optional[Union[timedelta, str]] = None,
             origin: Optional[shp.BaseGeometry] = None,
             start: Union[datetime, str, None] = None) -> \
             List[List[List[List[STBox]]]]:
        """
        Returns a 4D matrix (XxYxZxT) of `STBox` instances representing the
        tiles of ``self``. The resulting matrix has 4 dimensions regardless
        of the dimensionality of ``self``. If the ``self`` is missing a
        dimension, the resulting matrix will have a size of 1 for that dimension.

        Args:
            size: The size of the spatial tiles. If the `STBox` instance has a
                spatial dimension and this argument is not provided, the tiling
                will be only temporal.
            duration: The duration of the temporal tiles. If the `STBox`
                instance has a time dimension and this argument is not
                provided, the tiling will be only spatial.
            origin: The origin of the spatial tiling. If not provided, the
                origin will be (0, 0, 0).
            start: The start time of the temporal tiling. If not provided,
                the start time used by default is Monday, January 3, 2000.

        Returns:
            A 4D matrix (XxYxZxT) of `STBox` instances.

        MEOS Functions:
            stbox_tile_list
        """
        sz = size or (max(self.xmax() - self.xmin(), self.ymax() - self.ymin(),
                          (self.zmax() - self.zmin() if self.has_z() else 0)) + 1)
        dt = timedelta_to_interval(duration) if isinstance(duration, timedelta) \
            else pg_interval_in(duration, -1) if isinstance(duration, str) \
            else None
        st = datetime_to_timestamptz(start) if isinstance(start, datetime) \
            else pg_timestamptz_in(start, -1) if isinstance(start, str) \
            else pg_timestamptz_in('2000-01-03', -1) if self.has_t() \
            else 0
        gs = geo_to_gserialized(origin, self.geodetic()) if origin is not None \
            else pgis_geography_in('Point(0 0 0)', -1) if self.geodetic() \
            else pgis_geometry_in('Point(0 0 0)', -1)
        tiles, dimensions = stbox_tile_list(self._inner, sz, sz, sz, dt, gs, st)
        x_size = dimensions[0] or 1
        y_size = dimensions[1] or 1
        z_size = dimensions[2] or 1
        t_size = dimensions[3] or 1
        x_factor = y_size * z_size * t_size
        y_factor = z_size * t_size
        z_factor = t_size
        return [[[[STBox(_inner=tiles + x * x_factor + y * y_factor + z * z_factor + t) 
                for t in range(t_size)]
                for z in range(z_size)] for y in range(y_size)] for x in range(x_size)]

    def tile_flat(self, size: float, 
                  duration: Optional[Union[timedelta, str]] = None,
                  origin: Optional[shp.BaseGeometry] = None,
                  start: Union[datetime, str, None] = None) -> List[STBox]:
        """
        Returns a flat list of `STBox` instances representing the tiles of
        ``self``.

        Args:
            size: The size of the spatial tiles. If the `STBox` instance has a
                spatial dimension and this argument is not provided, the tiling
                will be only temporal.
            duration: The duration of the temporal tiles. If the `STBox`
                instance has a time dimension and this argument is not
                provided, the tiling will be only spatial.
            origin: The origin of the spatial tiling. If not provided, the
                origin will be (0, 0, 0).
            start: The start time of the temporal tiling. If not provided, the
                start time used by default is Monday, January 3, 2000.

        Returns:
            A flat list of `STBox` instances.

        MEOS Functions:
            stbox_tile_list
        """
        boxes = self.tile(size, duration, origin, start)
        return [b
                for x in boxes
                for y in x
                for z in y
                for b in z
                ]

    # ------------------------- Comparisons -----------------------------------
    def __eq__(self, other):
        """
        Returns whether ``self`` is equal to `other`.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is equal to ``other``, ``False`` otherwise.

        MEOS Functions:
            stbox_eq
        """
        if isinstance(other, self.__class__):
            return stbox_eq(self._inner, other._inner)
        return False

    def __ne__(self, other):
        """
        Returns whether ``self`` is not equal to `other`.

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is not equal to ``other``, ``False`` otherwise.

        MEOS Functions:
            stbox_ne
        """
        if isinstance(other, self.__class__):
            return stbox_ne(self._inner, other._inner)
        return True

    def __lt__(self, other):
        """
        Returns whether ``self`` is less than `other`. Compares first the SRID,
        then the time dimension, and finally the spatial dimension (X, then Y
        then Z lower bounds and then the upper bounds).

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is less than ``other``, ``False`` otherwise.

        MEOS Functions:
            stbox_lt
        """
        if isinstance(other, self.__class__):
            return stbox_lt(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __le__(self, other):
        """
        Returns whether ``self`` is less than or equal to `other`. Compares
        first the SRID, then the time dimension, and finally the spatial
        dimension (X, then Y then Z lower bounds and then the upper bounds).

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is less than or equal to ``other``, ``False``
            otherwise.

        MEOS Functions:
            stbox_le
        """
        if isinstance(other, self.__class__):
            return stbox_le(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __gt__(self, other):
        """
        Returns whether ``self`` is greater than `other`. Compares first the
        SRID, then the time dimension, and finally the spatial dimension (X,
        then Y then Z lower bounds and then the upper bounds).

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is greater than ``other``, ``False`` otherwise.

        MEOS Functions:
            stbox_gt
        """
        if isinstance(other, self.__class__):
            return stbox_gt(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __ge__(self, other):
        """
        Returns whether ``self`` is greater than or equal to `other`. Compares
        first the SRID, then the time dimension, and finally the spatial
        dimension (X, then Y then Z lower bounds and then the upper bounds).

        Args:
            other: The spatiotemporal object to compare with ``self``.

        Returns:
            ``True`` if ``self`` is greater than or equal to ``other``,
            ``False`` otherwise.

        MEOS Functions:
            stbox_ge
        """
        if isinstance(other, self.__class__):
            return stbox_ge(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    # ------------------------- Plot Operations -------------------------------
    def plot_xy(self, *args, **kwargs):
        """
        Plots the spatial dimension (XY) of ``self``.

        See Also:
            :func:`~pymeos.plotters.box_plotter.BoxPlotter.plot_stbox_xy`
        """
        from ..plotters import BoxPlotter
        return BoxPlotter.plot_stbox_xy(self, *args, **kwargs)

    def plot_xt(self, *args, **kwargs):
        """
        Plots the first spatial dimension and the temporal dimension (XT) of
        ``self``.

        See Also:
            :func:`~pymeos.plotters.box_plotter.BoxPlotter.plot_stbox_xt`
        """
        from ..plotters import BoxPlotter
        return BoxPlotter.plot_stbox_xt(self, *args, **kwargs)

    def plot_yt(self, *args, **kwargs):
        """
        Plots the second spatial dimension and the temporal dimension (YT) of
        ``self``.

        See Also:
            :func:`~pymeos.plotters.box_plotter.BoxPlotter.plot_stbox_yt`
        """
        from ..plotters import BoxPlotter
        return BoxPlotter.plot_stbox_yt(self, *args, **kwargs)

    # ------------------------- Database Operations ---------------------------
    @staticmethod
    def read_from_cursor(value, _=None):
        """
        Reads a :class:`STBox` from a database cursor. Used when automatically
        loading objects from the database.
        Users should use the class constructor instead.
        """
        if not value:
            return None
        return STBox(string=value)



from __future__ import annotations

from abc import ABC
from functools import reduce
from typing import Optional, List, Union, TYPE_CHECKING, Set, overload

from pymeos_cffi import *

from .tnumber import TNumber
from ..temporal import TInterpolation, Temporal, TInstant, TSequence, TSequenceSet
from ..collections import *

if TYPE_CHECKING:
    from ..boxes import TBox
    from .tint import TInt


class TFloat(TNumber[float, 'TFloat', 'TFloatInst', 'TFloatSeq', 'TFloatSeqSet'], ABC):
    _mobilitydb_name = 'tfloat'

    BaseClass = float
    _parse_function = tfloat_in

    # ------------------------- Constructors ----------------------------------
    @staticmethod
    def from_base_temporal(value: float, base: Temporal,
                           interpolation: TInterpolation = TInterpolation.LINEAR) -> TFloat:
        """
        Returns a new temporal float with the value `value` and the temporal
        frame of `base`.

        Args:
            value: Value of the temporal float.
            base: Temporal object to get the temporal frame from.
            interpolation: Interpolation of the temporal float.

        Returns:
            A new :class:`TFloat` object.

        MEOS Functions:
            tfloat_from_base_temp
        """
        result = tfloat_from_base_temp(value, base._inner)
        return Temporal._factory(result)

    @staticmethod
    @overload
    def from_base_time(value: float, base: datetime) -> TFloatInst:
        ...

    @staticmethod
    @overload
    def from_base_time(value: float, base: Union[TimestampSet, Period]) -> \
            TFloatSeq:
        ...

    @staticmethod
    @overload
    def from_base_time(value: float, base: PeriodSet) -> TFloatSeqSet:
        ...

    @staticmethod
    def from_base_time(value: float, base: Time,
                       interpolation: TInterpolation = None) -> TFloat:
        """
        Returns a new temporal float with the value `value` and the temporal
        frame of `base`.

        Args:
            value: Value of the temporal float.
            base: Time object to get the temporal frame from.
            interpolation: Interpolation of the temporal float.

        Returns:
            A new temporal float.

        MEOS Functions:
            tfloatinst_make, tfloatseq_from_base_timestampset,
            tfloatseq_from_base_time, tfloatseqset_from_base_time
        """
        if isinstance(base, datetime):
            return TFloatInst(_inner=tfloatinst_make(value,
                                                     datetime_to_timestamptz(base)))
        elif isinstance(base, TimestampSet):
            return TFloatSeq(_inner=tfloatseq_from_base_timestampset(value,
                                                                     base._inner))
        elif isinstance(base, Period):
            return TFloatSeq(_inner=tfloatseq_from_base_period(value,
                                                               base._inner, interpolation))
        elif isinstance(base, PeriodSet):
            return TFloatSeqSet(_inner=tfloatseqset_from_base_periodset(value,
                                                                        base._inner, interpolation))
        raise TypeError(f'Operation not supported with type {base.__class__}')

    # ------------------------- Output ----------------------------------------
    def __str__(self, max_decimals: int = 15):
        """
        Returns a string representation of `self`.

        Returns:
            A string representation of `self`.

        MEOS Functions:
            tfloat_out
        """
        return tfloat_out(self._inner, max_decimals)

    def as_wkt(self, max_decimals: int = 15) -> str:
        """
        Returns a WKT representation of `self`.

        Args:
            max_decimals: The number of decimals to use.

        Returns:
            A WKT representation of `self`.

        MEOS Functions:
            tfloat_out
        """
        return tfloat_out(self._inner, max_decimals)

    # ------------------------- Conversions ----------------------------------
    def to_tint(self) -> TInt:
        """
        Returns a new temporal integer with the values of `self` floored.
        This operation can only be performed when the interpolation is stepwise
        or discrete.

        Returns:
            A new temporal integer.

        MEOS Functions:
            tfloat_to_tint

        Raises:
            ValueError: If the interpolation is linear.
        """
        from ..factory import _TemporalFactory
        if self.interpolation() == TInterpolation.LINEAR:
            raise ValueError("Cannot convert a temporal float with linear " \
                             "interpolation to a temporal integer")
        return _TemporalFactory.create_temporal(tfloat_to_tint(self._inner))

    def to_floatrange(self) -> FloatSpan:
        """
        Returns value span of `self`.

        Returns:
            An :class:`FloatSpan` with the value span of `self`.

        MEOS Functions:
            tnumber_to_span
        """
        return FloatSpan(_inner=tnumber_to_span(self._inner))

    # ------------------------- Accessors -------------------------------------
    def value_span(self) -> FloatSpan:
        """
        Returns the value span of `self`.

        Returns:
            An :class:`FloatSpan` with the value span of `self`.

        MEOS Functions:
            tnumber_to_span
        """
        return self.to_floatrange()

    def value_spans(self) -> FloatSpanSet:
        """
        Returns the value spans of `self` taking into account gaps.

        Returns:
            A :class:`FloatSpanSet` with the value spans of `self`.

        MEOS Functions:
            tnumber_valuespans
        """
        return FloatSpanSet(_inner=(tnumber_valuespans(self._inner)))

    def start_value(self) -> float:
        """
        Returns the start value of `self`.

        Returns:
            A :class:`float` with the start value.

        MEOS Functions:
            tfloat_start_value
        """
        return tfloat_start_value(self._inner)

    def end_value(self) -> float:
        """
        Returns the end value of `self`.

        Returns:
            A :class:`float` with the end value.

        MEOS Functions:
            tfloat_end_value
        """
        return tfloat_end_value(self._inner)

    def value_set(self) -> Set[float]:
        """
        Returns the set of values of `self`.
        Note that when the interpolation is linear, the set will contain only
        the waypoints.

        Returns:
            A :class:`set` with the values of `self`.

        MEOS Functions:
            tint_values
        """
        values, count = tfloat_values(self._inner)
        return {values[i] for i in range(count)}

    def min_value(self) -> float:
        """
        Returns the minimum value of the `self`.

        Returns:
            A :class:`float` with the minimum value.

        MEOS Functions:
            tfloat_min_value
        """
        return tfloat_min_value(self._inner)

    def max_value(self) -> float:
        """
        Returns the maximum value of the `self`.

        Returns:
            A :class:`float` with the maximum value.

        MEOS Functions:
            tfloat_max_value
        """
        return tfloat_max_value(self._inner)

    # ------------------------- Ever and Always Comparisons -------------------
    def always_equal(self, value: float) -> bool:
        """
        Returns whether the values of `self` are always equal to `value`.

        Args:
            value: :class:`float` to compare.

        Returns:
            `True` if the values of `self` are always equal to `value`,
            `False` otherwise.

        MEOS Functions:
            tfloat_always_eq
        """
        return tfloat_always_eq(self._inner, value)

    def always_not_equal(self, value: float) -> bool:
        """
        Returns whether the values of `self` are always not equal to `value`.

        Args:
            value: :class:`float` to compare.

        Returns:
            `True` if the values of `self` are always not equal to `value`,
            `False` otherwise.

        MEOS Functions:
            tfloat_ever_eq
        """
        return not tfloat_ever_eq(self._inner, value)

    def always_less(self, value: float) -> bool:
        """
        Returns whether the values of `self` are always less than `value`.

        Args:
            value: :class:`float` to compare.

        Returns:
            `True` if the values of `self` are always less than `value`,
            `False` otherwise.

        MEOS Functions:
            tfloat_always_lt
        """
        return tfloat_always_lt(self._inner, value)

    def always_less_or_equal(self, value: float) -> bool:
        """
        Returns whether the values of `self` are always less than or equal to
        `value`.

        Args:
            value: :class:`float` to compare.

        Returns:
            `True` if the values of `self` are always less than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            tfloat_always_le
        """
        return tfloat_always_le(self._inner, value)

    def always_greater_or_equal(self, value: float) -> bool:
        """
        Returns whether the values of `self` are always greater than or equal
        to `value`.

        Args:
            value: :class:`float` to compare.

        Returns:
            `True` if the values of `self` are always greater than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            tfloat_ever_lt
        """
        return not tfloat_ever_lt(self._inner, value)

    def always_greater(self, value: float) -> bool:
        """
        Returns whether the values of `self` are always greater than `value`.

        Args:
            value: :class:`float` to compare.

        Returns:
            `True` if the values of `self` are always greater than `value`,
            `False` otherwise.

        MEOS Functions:
            tfloat_ever_le
        """
        return not tfloat_ever_le(self._inner, value)

    def ever_less(self, value: float) -> bool:
        """
        Returns whether the values of `self` are ever less than `value`.

        Args:
            value: :class:`float` to compare.

        Returns:
            `True` if the values of `self` are ever less than `value`,
            `False` otherwise.

        MEOS Functions:
            tfloat_ever_lt
        """
        return tfloat_ever_lt(self._inner, value)

    def ever_less_or_equal(self, value: float) -> bool:
        """
        Returns whether the values of `self` are ever less than or equal to
        `value`.

        Args:
            value: :class:`float` to compare.

        Returns:
            `True` if the values of `self` are ever less than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            tfloat_ever_le
        """
        return tfloat_ever_le(self._inner, value)

    def ever_equal(self, value: float) -> bool:
        """
        Returns whether the values of `self` are ever equal to `value`.

        Args:
            value: :class:`float` to compare.

        Returns:
            `True` if the values of `self` are ever equal to `value`, `False`
            otherwise.

        MEOS Functions:
            tfloat_ever_eq
        """
        return tfloat_ever_eq(self._inner, value)

    def ever_not_equal(self, value: float) -> bool:
        """
        Returns whether the values of `self` are ever not equal to `value`.

        Args:
            value: :class:`float` to compare.

        Returns:
            `True` if the values of `self` are ever not equal to `value`,
            `False` otherwise.

        MEOS Functions:
            tfloat_always_eq
        """
        return not tfloat_always_eq(self._inner, value)

    def ever_greater_or_equal(self, value: float) -> bool:
        """
        Returns whether the values of `self` are ever greater than or equal to
        `value`.

        Args:
            value: :class:`float` to compare.

        Returns:
            `True` if the values of `self` are ever greater than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            tfloat_always_lt
        """
        return not tfloat_always_lt(self._inner, value)

    def ever_greater(self, value: float) -> bool:
        """
        Returns whether the values of `self` are ever greater than `value`.

        Args:
            value: :class:`float` to compare.

        Returns:
            `True` if the values of `self` are ever greater than `value`,
            `False` otherwise.

        MEOS Functions:
            tfloat_always_le
        """
        return not tfloat_always_le(self._inner, value)

    def never_equal(self, value: float) -> bool:
        """
        Returns whether the values of `self` are never equal to `value`.

        Args:
            value: :class:`float` value to compare.

        Returns:
            `True` if the values of `self` are never equal to `value`,
            `False` otherwise.

        MEOS Functions:
            tfloat_ever_eq
        """
        return not tfloat_ever_eq(self._inner, value)

    def never_not_equal(self, value: float) -> bool:
        """
        Returns whether the values of `self` are never not equal to `value`.

        Args:
            value: :class:`float` value to compare.

        Returns:
            `True` if the values of `self` are never not equal to `value`,
            `False` otherwise.

        MEOS Functions:
            tfloat_always_eq
        """
        return tfloat_always_eq(self._inner, value)

    def never_less(self, value: float) -> bool:
        """
        Returns whether the values of `self` are never less than `value`.

        Args:
            value: :class:`float` value to compare.

        Returns:
            `True` if the values of `self` are never less than `value`,
            `False` otherwise.

        MEOS Functions:
            tfloat_ever_lt
        """
        return not tfloat_ever_lt(self._inner, value)

    def never_less_or_equal(self, value: float) -> bool:
        """
        Returns whether the values of `self` are never less than or equal to
        `value`.

        Args:
            value: :class:`float` value to compare.

        Returns:
            `True` if the values of `self` are never less than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            tfloat_ever_le
        """
        return not tfloat_ever_le(self._inner, value)

    def never_greater_or_equal(self, value: float) -> bool:
        """
        Returns whether the values of `self` are never greater than or equal to
        `value`.

        Args:
            value: :class:`float` value to compare.

        Returns:
            `True` if the values of `self` are never greater than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            tfloat_always_lt
        """
        return tfloat_always_lt(self._inner, value)

    def never_greater(self, value: float) -> bool:
        """
        Returns whether the values of `self` are never greater than `value`.

        Args:
            value: :class:`float` value to compare.

        Returns:
            `True` if the values of `self` are never greater than `value`,
            `False` otherwise.

        MEOS Functions:
            tfloat_always_le
        """
        return tfloat_always_le(self._inner, value)

    # ------------------------- Temporal Comparisons --------------------------
    def temporal_equal(self, other: Union[int, float, Temporal]) -> Temporal:
        """
        Returns the temporal equality relation between `self` and `other`.

        Args:
            other: An :class:`int`, :class:`float` or temporal object to
            compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal equality relation.

        MEOS Functions:
            teq_tfloat_float, teq_temporal_temporal
        """
        if isinstance(other, int) or isinstance(other, float):
            result = teq_tfloat_float(self._inner, float(other))
        else:
            return super().temporal_equal(other)
        return Temporal._factory(result)

    def temporal_not_equal(self, other: Union[int, float, Temporal]) -> Temporal:
        """
        Returns the temporal not equal relation between `self` and `other`.

        Args:
            other: An :class:`int`, :class:`float` or temporal object to
            compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal not equal relation.

        MEOS Functions:
            tne_tfloat_float, tne_temporal_temporal
        """
        if isinstance(other, int) or isinstance(other, float):
            result = tne_tfloat_float(self._inner, float(other))
        else:
            return super().temporal_not_equal(other)
        return Temporal._factory(result)

    def temporal_less(self, other: Union[int, float, Temporal]) -> Temporal:
        """
        Returns the temporal less than relation between `self` and `other`.

        Args:
            other: An :class:`int`, :class:`float` or temporal object to
            compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal less than relation.

        MEOS Functions:
            tlt_tfloat_float, tlt_temporal_temporal
        """
        if isinstance(other, int) or isinstance(other, float):
            result = tlt_tfloat_float(self._inner, float(other))
        else:
            return super().temporal_less(other)
        return Temporal._factory(result)

    def temporal_less_or_equal(self, other: Union[int, float, Temporal]) -> \
            Temporal:
        """
        Returns the temporal less or equal relation between `self` and `other`.

        Args:
            other: An :class:`int`, :class:`float` or temporal object to
            compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal less or equal
            relation.

        MEOS Functions:
            tle_tfloat_float, tle_temporal_temporal
        """
        if isinstance(other, int) or isinstance(other, float):
            result = tle_tfloat_float(self._inner, float(other))
        else:
            return super().temporal_less_or_equal(other)
        return Temporal._factory(result)

    def temporal_greater_or_equal(self, other: Union[int, float, Temporal]) \
            -> Temporal:
        """
        Returns the temporal greater or equal relation between `self` and `other`.

        Args:
            other: An :class:`int`, :class:`float` or temporal object to
            compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal greater or equal
            relation.

        MEOS Functions:
            tge_tfloat_float, tge_temporal_temporal
        """
        if isinstance(other, int) or isinstance(other, float):
            result = tge_tfloat_float(self._inner, float(other))
        else:
            return super().temporal_greater_or_equal(other)
        return Temporal._factory(result)

    def temporal_greater(self, other: Union[int, float, Temporal]) -> Temporal:
        """
        Returns the temporal greater than relation between `self` and `other`.

        Args:
            other: An :class:`int`, :class:`float` or temporal object to
            compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal greater than
            relation.

        MEOS Functions:
            tgt_tfloat_float, tgt_temporal_temporal
        """
        if isinstance(other, int) or isinstance(other, float):
            result = tgt_tfloat_float(self._inner, float(other))
        else:
            return super().temporal_greater(other)
        return Temporal._factory(result)

    # ------------------------- Restrictions ----------------------------------
    def at(self, other: Union[float, int, FloatSet, IntSet, FloatSpan, IntSpan, \
            FloatSpanSet, IntSpanSet, TBox, Time]) -> TFloat:
        """
        Returns a new temporal float with the values of `self` restricted to
        the value or time `other`.

        Args:
            other: Value or time to restrict to.

        Returns:
            A new temporal float.

        MEOS Functions:
            tfloat_at_value, temporal_at_values, tnumber_at_span, tnumber_at_spanset,
            temporal_at_timestamp, temporal_at_timestampset, temporal_at_period,
            temporal_at_periodset
        """
        if isinstance(other, int) or isinstance(other, float):
            result = tfloat_at_value(self._inner, float(other))
        elif isinstance(other, IntSet):
            return super().at(other.to_floatset())
        elif isinstance(other, IntSpan):
            return super().at(other.to_floatspan())
        elif isinstance(other, IntSpanSet):
            return super().at(other.to_floatspanset())
        else:
            return super().at(other)
        return Temporal._factory(result)

    def minus(self, other: Union[float, int, FloatSet, IntSet, FloatSpan, IntSpan, \
            FloatSpanSet, IntSpanSet, TBox, Time]) -> Temporal:
        """
        Returns a new temporal float with the values of `self` restricted to
        the complement of the time or value `other`.

        Args:
            other: Time or value to restrict to the complement of.

        Returns:
            A new temporal float.

        MEOS Functions:
            tfloat_minus_value, temporal_minus_values, tnumber_minus_span, tnumber_minus_spanset,
            temporal_minus_timestamp, temporal_minus_timestampset,
            temporal_minus_period, temporal_minus_periodset
        """
        if isinstance(other, int) or isinstance(other, float):
            result = tfloat_minus_value(self._inner, float(other))
        elif isinstance(other, IntSet):
            return super().minus(other.to_floatset())
        elif isinstance(other, IntSpan):
            return super().minus(other.to_floatspan())
        elif isinstance(other, IntSpanSet):
            return super().minus(other.to_floatspanset())
        else:
            return super().minus(other)
        return Temporal._factory(result)

    def value_at_timestamp(self, timestamp) -> float:
        """
        Returns the value that `self` takes at a certain moment.

        Args:
            timestamp: The moment to get the value.

        Returns:
            A class:`float` with the value of `self` at `timestamp`.

        MEOS Functions:
            tfloat_value_at_timestamp
        """
        return tfloat_value_at_timestamp(self._inner,
                                         datetime_to_timestamptz(timestamp), True)

    def derivative(self) -> TFloat:
        """
        Returns the derivative of `self`.

        Returns:
            A new :class:`TFloat` instance.

        MEOS Functions:
            tfloat_derivative
        """
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(tfloat_derivative(self._inner))

    # ------------------------- Transformations ----------------------------------
    def to_degrees(self, normalize: bool = True) -> TFloat:
        """
        Returns a copy of `self` converted from radians to degrees.

        Args:
            normalize: If True, the result will be normalized to the range
            [0, 360).

        Returns:
            A new :class:`TFloat` instance.

        MEOS Functions:
            tfloat_degrees
        """
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(tfloat_degrees(self._inner,
                                                               normalize))

    def to_radians(self) -> TFloat:
        """
        Returns a copy of `self` converted from degrees to radians.

        Returns:
            A new :class:`TFloat` instance.

        MEOS Functions:
            tfloat_radians
        """
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(tfloat_radians(self._inner))

    def round(self, maxdd: int = 0) -> TFloat:
        """
        Returns `self` rounded to the given number of decimal digits.

        Args:
            maxdd: Maximum number of decimal digits.

        Returns:
            A new :class:`TFloat` instance.

        MEOS Functions:
            tfloat_round
        """
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(tfloat_round(self._inner,
                                                             maxdd))

    # ------------------------- Split Operations ------------------------------
    def value_split(self, size: float, start: Optional[float] = 0) -> \
            List[Temporal]:
        """
        Splits `self` into fragments with respect to value buckets

        Args:
            start: Start value of the first value bucket.
            size: Size of the value buckets.

        Returns:
            A list of temporal floats.

        MEOS Functions:
            tfloat_value_split
        """
        fragments, _, count = tfloat_value_split(self._inner, size, start)
        from ..factory import _TemporalFactory
        return [_TemporalFactory.create_temporal(fragments[i]) for i in \
                range(count)]

    def value_time_split(self, value_size: float,
                         duration: Union[str, timedelta],
                         value_start: Optional[float] = 0.0,
                         time_start: Optional[Union[str, datetime]] = None) -> \
            List[Temporal]:
        """
        Splits `self` into fragments with respect to value and period buckets.

        Args:
            value_size: Size of the value buckets.
            duration: Duration of the period buckets.
            value_start: Start value of the first value bucket. If None, the
                start value used by default is 0
            time_start: Start time of the first period bucket. If None, the
                start time used by default is Monday, January 3, 2000.

        Returns:
            A list of temporal floats.

        MEOS Functions:
            tfloat_value_time_split
        """
        if time_start is None:
            st = pg_timestamptz_in('2000-01-03', -1)
        else:
            st = datetime_to_timestamptz(time_start) \
                if isinstance(time_start, datetime) \
                else pg_timestamptz_in(time_start, -1)
        dt = timedelta_to_interval(duration) \
            if isinstance(duration, timedelta) \
            else pg_interval_in(duration, -1)
        fragments, _, _, count = tfloat_value_time_split(self._inner,
                                                         value_size, dt, value_start, st)
        return [Temporal._factory(fragments[i]) for i in range(count)]

    # ------------------------- Database Operations ---------------------------
    @staticmethod
    def read_from_cursor(value, _=None):
        """
        Reads a :class:`TFloat` from a database cursor. Used when automatically
        loading objects from the database.
        Users should use the class constructor instead.
        """
        if not value:
            return None
        if value.startswith('Interp=Stepwise;'):
            value1 = value.replace('Interp=Stepwise;', '')
            if value1[0] == '{':
                return TFloatSeqSet(string=value)
            else:
                return TFloatSeq(string=value)
        elif value[0] != '{' and value[0] != '[' and value[0] != '(':
            return TFloatInst(string=value)
        elif value[0] == '[' or value[0] == '(':
            return TFloatSeq(string=value)
        elif value[0] == '{':
            if value[1] == '[' or value[1] == '(':
                return TFloatSeqSet(string=value)
            else:
                return TFloatSeq(string=value)
        raise Exception("ERROR: Could not parse temporal float value")


class TFloatInst(TInstant[float, 'TFloat', 'TFloatInst', 'TFloatSeq',
'TFloatSeqSet'], TFloat):
    """
    Class for representing temporal floats at a single instant.
    """
    _make_function = tfloatinst_make
    _cast_function = float

    def __init__(self, string: Optional[str] = None, *,
                 value: Optional[Union[str, float]] = None,
                 timestamp: Optional[Union[str, datetime]] = None, _inner=None):
        super().__init__(string=string, value=value, timestamp=timestamp,
                         _inner=_inner)


class TFloatSeq(TSequence[float, 'TFloat', 'TFloatInst', 'TFloatSeq',
'TFloatSeqSet'], TFloat):
    """
    Class for representing temporal floats over a period of time.
    """
    ComponentClass = TFloatInst

    def __init__(self, string: Optional[str] = None, *,
                 instant_list: Optional[List[Union[str, TFloatInst]]] = None,
                 lower_inc: bool = True, upper_inc: bool = False,
                 interpolation: TInterpolation = TInterpolation.LINEAR,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, instant_list=instant_list,
                         lower_inc=lower_inc, upper_inc=upper_inc,
                         interpolation=interpolation,
                         normalize=normalize, _inner=_inner)


class TFloatSeqSet(TSequenceSet[float, 'TFloat', 'TFloatInst', 'TFloatSeq',
'TFloatSeqSet'], TFloat):
    """
    Class for representing temporal floats over a period of time with gaps.
    """
    ComponentClass = TFloatSeq

    def __init__(self, string: Optional[str] = None, *,
                 sequence_list: Optional[List[Union[str, TFloatSeq]]] = None,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, sequence_list=sequence_list,
                         normalize=normalize, _inner=_inner)



from __future__ import annotations

from abc import ABC
from functools import reduce
from typing import Optional, Union, List, Set, overload, TYPE_CHECKING

from pymeos_cffi import *

from ..temporal import TInterpolation, Temporal, TInstant, TSequence, TSequenceSet
from ..collections import *

if TYPE_CHECKING:
    from .tbool import TBool


class TText(Temporal[str, 'TText', 'TTextInst', 'TTextSeq', 'TTextSeqSet'], ABC):
    _mobilitydb_name = 'ttext'

    BaseClass = str

    _parse_function = ttext_in

    def __init__(self, _inner) -> None:
        super().__init__()

    # ------------------------- Constructors ----------------------------------
    @staticmethod
    def from_base_temporal(value: str, base: Temporal) -> TText:
        """
        Create a temporal string from a string and the time frame of another
        temporal object.

        Args:
            value: string value of the temporal.
            base: Temporal object to use as time frame.

        Returns:
            A new :class:`TText` object.

        MEOS Functions:
            ttext_from_base_temp
        """
        result = ttext_from_base_temp(value, base._inner)
        return Temporal._factory(result)

    @staticmethod
    @overload
    def from_base_time(value: str, base: datetime) -> TTextInst:
        ...

    @staticmethod
    @overload
    def from_base_time(value: str, base: Union[TimestampSet, Period]) -> \
            TTextSeq:
        ...

    @staticmethod
    @overload
    def from_base_time(value: str, base: PeriodSet) -> TTextSeqSet:
        ...

    @staticmethod
    def from_base_time(value: str, base: Time) -> TText:
        """
        Create a temporal string from a :class:`str` and a time object.

        Args:
            value: string value.
            base: Time object to use as temporal dimension.

        Returns:
            A new temporal string.

        MEOS Functions:
            ttextinst_make, ttextseq_from_base_timestampset,
            ttextseq_from_base_period, ttextseqset_from_base_periodset
        """
        if isinstance(base, datetime):
            return TTextInst(_inner=ttextinst_make(value,
                                                   datetime_to_timestamptz(base)))
        elif isinstance(base, TimestampSet):
            return TTextSeq(_inner=ttextseq_from_base_timestampset(value,
                                                                   base._inner))
        elif isinstance(base, Period):
            return TTextSeq(_inner=ttextseq_from_base_period(value,
                                                             base._inner))
        elif isinstance(base, PeriodSet):
            return TTextSeqSet(_inner=ttextseqset_from_base_periodset(value,
                                                                      base._inner))
        raise TypeError(f'Operation not supported with type {base.__class__}')

    # ------------------------- Output ----------------------------------------
    def __str__(self) -> str:
        """
        Returns the string representation of `self`.

        Returns:
            A string with the string representation of `self`.

        MEOS Functions:
            ttext_out
        """
        return ttext_out(self._inner)

    def as_wkt(self) -> str:
        """
        Returns the Well-Known Text representation of `self`.

        Returns:
            A string with the Well-Known Text representation of `self`.

        MEOS Functions:
            ttext_as_wkt
        """
        return ttext_out(self._inner)

    # ------------------------- Accessors -------------------------------------
    def value_set(self) -> Set[str]:
        """
        Returns the set of unique values of the temporal string.

        Returns:
            A set of unique values.

        MEOS Functions:
            ttext_values
        """
        values, count = ttext_values(self._inner)
        return {text2cstring(values[i]) for i in range(count)}

    def min_value(self) -> str:
        """
        Returns the minimum value of the temporal string.

        Returns:
            A :class:`str` with the minimum value.

        MEOS Functions:
            ttext_min_value
        """
        return ttext_min_value(self._inner)

    def max_value(self) -> str:
        """
        Returns the maximum value of the temporal string.

        Returns:
            A :class:`str` with the maximum value.

        MEOS Functions:
            ttext_max_value
        """
        return ttext_max_value(self._inner)

    def start_value(self) -> str:
        """
        Returns the start value of the temporal string.

        Returns:
            A :class:`str` with the start value.

        MEOS Functions:
            ttext_start_value
        """
        return ttext_start_value(self._inner)

    def end_value(self) -> str:
        """
        Returns the end value of the temporal string.

        Returns:
            A :class:`str` with the end value.

        MEOS Functions:
            ttext_end_value
        """
        return ttext_end_value(self._inner)

    def upper(self) -> TText:
        """
        Returns a new temporal string with the values of `self` converted to
        upper case.

        Returns:
            A new temporal string.

        MEOS Functions:
            ttext_upper
        """
        return self.__class__(_inner=ttext_upper(self._inner))

    def lower(self) -> TText:
        """
        Returns a new temporal string with the values of `self` converted to
        lower case.

        Returns:
            A new temporal string.

        MEOS Functions:
            ttext_lower
        """
        return self.__class__(_inner=ttext_lower(self._inner))

    def value_at_timestamp(self, timestamp: datetime) -> str:
        """
        Returns the value that `self` takes at a certain moment.

        Args:
            timestamp: The moment to get the value.

        Returns:
            A string with the value of `self` at `timestamp`.

        MEOS Functions:
            ttext_value_at_timestamp
        """
        result = ttext_value_at_timestamp(self._inner,
                                          datetime_to_timestamptz(timestamp), True)
        return text2cstring(result[0])

    # ------------------------- Ever and Always Comparisons -------------------
    def always_equal(self, value: str) -> bool:
        """
        Returns whether the values of `self` are always equal to `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are always equal to `value`,
            `False` otherwise.

        MEOS Functions:
            ttext_always_eq
        """
        return ttext_always_eq(self._inner, value)

    def always_not_equal(self, value: str) -> bool:
        """
        Returns whether the values of `self` are always not equal to `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are always not equal to `value`,
            `False` otherwise.

        MEOS Functions:
            ttext_ever_eq
        """
        return not ttext_always_eq(self._inner, value)

    def always_less(self, value: str) -> bool:
        """
        Returns whether the values of `self` are always less than `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are always less than `value`,
            `False` otherwise.

        MEOS Functions:
            ttext_always_lt
        """
        return ttext_always_lt(self._inner, value)

    def always_less_or_equal(self, value: str) -> bool:
        """
        Returns whether the values of `self` are always less than or equal to
        `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are always less than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            ttext_always_le
        """
        return ttext_always_le(self._inner, value)

    def always_greater(self, value: str) -> bool:
        """
        Returns whether the values of `self` are always greater than `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are always greater than `value`,
            `False` otherwise.

        MEOS Functions:
            ttext_ever_le
        """
        return not ttext_ever_le(self._inner, value)

    def always_greater_or_equal(self, value: str) -> bool:
        """
        Returns whether the values of `self` are always greater than or equal
        to `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are always greater than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            ttext_ever_lt
        """
        return not ttext_ever_lt(self._inner, value)

    def ever_equal(self, value: str) -> bool:
        """
        Returns whether the values of `self` are ever equal to `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are ever equal to `value`, `False`
            otherwise.

        MEOS Functions:
            ttext_ever_eq
        """
        return ttext_ever_eq(self._inner, value)

    def ever_not_equal(self, value: str) -> bool:
        """
        Returns whether the values of `self` are ever not equal to `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are ever not equal to `value`,
            `False` otherwise.

        MEOS Functions:
            ttext_always_eq
        """
        return not ttext_ever_eq(self._inner, value)

    def ever_less(self, value: str) -> bool:
        """
        Returns whether the values of `self` are ever less than `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are ever less than `value`, `False`
            otherwise.

        MEOS Functions:
            ttext_ever_lt
        """
        return ttext_ever_lt(self._inner, value)

    def ever_less_or_equal(self, value: str) -> bool:
        """
        Returns whether the values of `self` are ever less than or equal to
        `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are ever less than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            ttext_ever_le
        """
        return ttext_ever_le(self._inner, value)

    def ever_greater(self, value: str) -> bool:
        """
        Returns whether the values of `self` are ever greater than `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are ever greater than `value`,
            `False` otherwise.

        MEOS Functions:
            ttext_always_le
        """
        return not ttext_always_le(self._inner, value)

    def ever_greater_or_equal(self, value: str) -> bool:
        """
        Returns whether the values of `self` are ever greater than or equal to
        `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are ever greater than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            ttext_always_lt
        """
        return not ttext_always_lt(self._inner, value)

    def never_equal(self, value: str) -> bool:
        """
        Returns whether the values of `self` are never equal to `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are never equal to `value`, `False`
            otherwise.

        MEOS Functions:
            ttext_ever_eq
        """
        return not ttext_ever_eq(self._inner, value)

    def never_not_equal(self, value: str) -> bool:
        """
        Returns whether the values of `self` are never not equal to `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are never not equal to `value`,
            `False` otherwise.

        MEOS Functions:
            ttext_always_eq
        """
        return ttext_always_eq(self._inner, value)

    def never_less(self, value: str) -> bool:
        """
        Returns whether the values of `self` are never less than `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are never less than `value`, `False`
            otherwise.

        MEOS Functions:
            ttext_ever_lt
        """
        return not ttext_ever_lt(self._inner, value)

    def never_less_or_equal(self, value: str) -> bool:
        """
        Returns whether the values of `self` are never less than or equal to
        `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are never less than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            ttext_ever_le
        """
        return not ttext_ever_le(self._inner, value)

    def never_greater_or_equal(self, value: str) -> bool:
        """
        Returns whether the values of `self` are never greater than or equal to
        `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are never greater than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            ttext_always_lt
        """
        return ttext_always_lt(self._inner, value)

    def never_greater(self, value: str) -> bool:
        """
        Returns whether the values of `self` are never greater than `value`.

        Args:
            value: String value to compare.

        Returns:
            `True` if the values of `self` are never greater than `value`,
            `False` otherwise.

        MEOS Functions:
            ttext_always_le
        """
        return ttext_always_le(self._inner, value)

    # ------------------------- Temporal Comparisons --------------------------
    def temporal_equal(self, other: Union[str, Temporal]) -> Temporal:
        """
        Returns the temporal equality relation between `self` and `other`.

        Args:
            other: A string or temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal equality relation.

        MEOS Functions:
            teq_ttext_text, teq_temporal_temporal
        """
        if isinstance(other, str):
            result = teq_ttext_text(self._inner, other)
        else:
            return super().temporal_equal(other)
        return Temporal._factory(result)

    def temporal_not_equal(self, other: Union[str, Temporal]) -> Temporal:
        """
        Returns the temporal not equal relation between `self` and `other`.

        Args:
            other: A string or temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal not equal relation.

        MEOS Functions:
            tne_ttext_text, tne_temporal_temporal
        """
        if isinstance(other, str):
            result = tne_ttext_text(self._inner, other)
        else:
            return super().temporal_not_equal(other)
        return Temporal._factory(result)

    def temporal_less(self, other: Union[str, Temporal]) -> TBool:
        """
        Returns the temporal less than relation between `self` and `other`.

        Args:
            other: A string or temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal less than relation.

        MEOS Functions:
            tlt_ttext_text, tlt_temporal_temporal
        """
        if isinstance(other, str):
            result = tlt_ttext_text(self._inner, other)
        else:
            return super().temporal_less(other)
        return Temporal._factory(result)

    def temporal_less_or_equal(self, other: Union[str, Temporal]) -> Temporal:
        """
        Returns the temporal less or equal relation between `self` and `other`.

        Args:
            other: A string or temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal less or equal
            relation.

        MEOS Functions:
            tle_ttext_text, tle_temporal_temporal
        """
        if isinstance(other, str):
            result = tle_ttext_text(self._inner, other)
        else:
            return super().temporal_less_or_equal(other)
        return Temporal._factory(result)

    def temporal_greater(self, other: Union[str, Temporal]) -> Temporal:
        """
        Returns the temporal greater than relation between `self` and `other`.

        Args:
            other: A string or temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal greater than
            relation.

        MEOS Functions:
            tgt_ttext_text, tgt_temporal_temporal
        """
        if isinstance(other, str):
            result = tgt_ttext_text(self._inner, other)
        else:
            return super().temporal_greater(other)
        return Temporal._factory(result)

    def temporal_greater_or_equal(self, other: Union[str, Temporal]) -> Temporal:
        """
        Returns the temporal greater or equal relation between `self` and
        `other`.

        Args:
            other: A string or temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal greater or equal
            relation.

        MEOS Functions:
            tge_ttext_text, tge_temporal_temporal
        """
        if isinstance(other, str):
            result = tge_ttext_text(self._inner, other)
        else:
            return super().temporal_greater_or_equal(other)
        return Temporal._factory(result)

    # ------------------------- Restrictions ----------------------------------
    def at(self, other: Union[str, List[str], datetime, TimestampSet, Period,
    PeriodSet]) -> TText:
        """
        Returns a new temporal string with the values of `self` restricted to
        the time or value `other`.

        Args:
            other: Time or value to restrict to.

        Returns:
            A new temporal string.

        MEOS Functions:
            ttext_at_value, temporal_at_timestamp, temporal_at_timestampset,
            temporal_at_period, temporal_at_periodset
        """
        if isinstance(other, str):
            result = ttext_at_value(self._inner, other)
        elif isinstance(other, list) and isinstance(other[0], str):
            result = temporal_at_values(self._inner, textset_make(other))
        else:
            return super().at(other)
        return Temporal._factory(result)

    def minus(self, other: Union[str, List[str], datetime, TimestampSet,
    Period, PeriodSet]) -> TText:
        """
        Returns a new temporal string with the values of `self` restricted to
        the complement of the time or value `other`.

        Args:
            other: Time or value to restrict to the complement of.

        Returns:
            A new temporal string.

        MEOS Functions:
            ttext_minus_value, temporal_minus_timestamp,
            temporal_minus_timestampset, temporal_minus_period,
            temporal_minus_periodset
        """
        if isinstance(other, str):
            result = ttext_minus_value(self._inner, other)
        elif isinstance(other, list) and isinstance(other[0], str):
            result = temporal_minus_values(self._inner, textset_make(other))
        else:
            return super().minus(other)
        return Temporal._factory(result)

    # ------------------------- Text Operations ------------------------------
    def concatenate(self, other: Union[str, TText], other_before: bool = False):
        """
        Returns a new temporal string with the values of `self` concatenated
        with the values of `other`.

        Args:
            other: Temporal string or string to concatenate.
            other_before: If `True` the values of `other` are prepended to the
            values of `self`.

        Returns:
            A new temporal string.

        MEOS Functions:
            textcat_ttext_text, textcat_text_ttext, textcat_ttext_ttext

        """
        if isinstance(other, str):
            result = textcat_ttext_text(self._inner, other) \
                if not other_before \
                else textcat_text_ttext(other, self._inner)
        elif isinstance(other, TText):
            result = textcat_ttext_ttext(self._inner, other._inner) \
                if not other_before \
                else textcat_ttext_ttext(other._inner, self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return self.__class__(_inner=result)

    def __add__(self, other):
        """
        Returns a new temporal string with the values of `self` concatenated
        with the values of `other`.

        Args:
            other: Temporal string or string to concatenate.

        Returns:
            A new temporal string.

        MEOS Functions:
            textcat_ttext_text, textcat_text_ttext, textcat_ttext_ttext
        """
        return self.concatenate(other)

    def __radd__(self, other):
        """
        Returns a new temporal string with the values of `other` concatenated
        with the values of `self`.

        Args:
            other: Temporal string or string to concatenate.

        Returns:
            A new temporal string.

        MEOS Functions:
            textcat_ttext_text, textcat_text_ttext, textcat_ttext_ttext
        """
        return self.concatenate(other, True)

    # ------------------------- Database Operations ---------------------------
    @staticmethod
    def read_from_cursor(value, _=None):
        """
        Reads a :class:`TText` from a database cursor. Used when automatically
        loading objects from the database.
        Users should use the class constructor instead.
        """
        if not value:
            return None
        if value[0] != '{' and value[0] != '[' and value[0] != '(':
            return TTextInst(string=value)
        elif value[0] == '[' or value[0] == '(':
            return TTextSeq(string=value)
        elif value[0] == '{':
            if value[1] == '[' or value[1] == '(':
                return TTextSeqSet(string=value)
            else:
                return TTextSeq(string=value)
        raise Exception("ERROR: Could not parse temporal text value")


class TTextInst(TInstant[str, 'TText', 'TTextInst', 'TTextSeq',
'TTextSeqSet'], TText):
    """
    Class for representing temporal strings at a single instant.
    """
    _make_function = ttextinst_make
    _cast_function = str

    def __init__(self, string: Optional[str] = None, *,
                 value: Optional[str] = None,
                 timestamp: Optional[Union[str, datetime]] = None,
                 _inner=None):
        super().__init__(string=string, value=value, timestamp=timestamp,
                         _inner=_inner)


class TTextSeq(TSequence[str, 'TText', 'TTextInst', 'TTextSeq',
'TTextSeqSet'], TText):
    """
    Class for representing temporal strings over a period of time.
    """
    ComponentClass = TTextInst

    def __init__(self, string: Optional[str] = None, *,
                 instant_list: Optional[List[Union[str, TTextInst]]] = None,
                 lower_inc: bool = True, upper_inc: bool = False,
                 interpolation: TInterpolation = TInterpolation.STEPWISE,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, instant_list=instant_list,
                         lower_inc=lower_inc, upper_inc=upper_inc,
                         interpolation=interpolation,
                         normalize=normalize, _inner=_inner)


class TTextSeqSet(TSequenceSet[str, 'TText', 'TTextInst', 'TTextSeq',
'TTextSeqSet'], TText):
    """
    Class for representing temporal strings over a period of time with gaps.
    """
    ComponentClass = TTextSeq

    def __init__(self, string: Optional[str] = None, *,
                 sequence_list: Optional[List[Union[str, TTextSeq]]] = None,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, sequence_list=sequence_list,
                         normalize=normalize, _inner=_inner)
